r"""Custom samplers for ``torch.utils.data.DataLoader``."""

import math
import multiprocessing as mp
from typing import Iterator, Optional, Sequence

import torch
import torch.distributed as dist

from torch.utils.data.sampler import Sampler, T_co


class DistributedWeightedRandomSampler(Sampler[T_co]):
    r"""A version of WeightedRandomSampler that can be used with DistributedDataParallel training."""

    def __init__(
        self,
        weights: Sequence[float],
        num_samples: int = -1,
        replacement: bool = True,
        num_replicas: Optional[int] = None,
        rank: Optional[int] = None,
        shuffle: bool = True,
        seed: int = 0,
        drop_last: bool = False,
    ) -> None:
        r"""Initialize map-style dataset index sampler.

        Sampler that restricts data loading to a subset of the random samples.
        It is especially useful in conjunction with ``torch.nn.parallel.DistributedDataParallel``.
        In such a case, each process can pass a ``torch.utils.data.DistributedWeightedRandomSampler``
        instance as a ``torch.utils.data.DataLoader`` sampler, and load a subset of the
        original dataset that is exclusive to it.

        .. note::
            Dataset is assumed to be a map-style dataset of constant size.

        Args:
            weights: Sampling weights for each of the ``len(weights)`` indices.
                These are passed on to ``torch.multinomial()`` to sample indices.
            num_samples: Number of dataset indices to sample. If negative, use ``len(weights)``.
                Note that using all samples would only result in shuffling the dataset if
                ``replacement=False``, but no weighted sampling of a subset of it.
            replacement: Whether to sample dataset indices with replacement.
            num_replicas: Number of processes participating in distributed training.
                By default, ``world_size`` is retrieved from the current distributed group.
            rank: Rank of the current process within ``num_replicas``. By default, ``rank`` is
                retrieved from the current distributed group.
            shuffle: If ``True``, the current ``epoch`` is added to the ``seed`` value to draw
                different samples at each epoch. Moreover, when sampling without replacement,
                the sampler will further shuffle the randomly drawn indices.
            seed: Random seed used to draw and shuffle random samples. This number should be identical
                across all processes in the distributed group.
            drop_last: If ``True``, then the sampler will drop the tail of the data to make it evenly
                divisible across the number of replicas. If ``False``, the sampler will add extra indices
                to make the data evenly divisible across the replicas.

        .. warning::
            In distributed mode, calling the :meth:``set_epoch`` method at the beginning of each epoch
            **before** creating the ``DataLoader`` iterator is necessary to make shuffling work properly
            across multiple epochs. Otherwise, the same ordering will always be used. It should further
            be noted, that the ``epoch`` is stored in shared memory such that persistent worker processes
            all receive an update of the epoch number when ``set_epoch`` is called.

        """
        if num_replicas is None:
            if not dist.is_available():
                raise RuntimeError(
                    f"{type(self).__name__}() requires distributed package to be available"
                )
            num_replicas = dist.get_world_size()
        if rank is None:
            if not dist.is_available():
                raise RuntimeError(
                    f"{type(self).__name__}() requires distributed package to be available"
                )
            rank = dist.get_rank()
        if rank >= num_replicas or rank < 0:
            raise ValueError(
                f"{type(self).__name__}() invalid rank {rank}"
                f", rank should be in the interval [0, {num_replicas - 1}]"
            )
        if num_samples < 1:
            num_samples = len(weights)
        self.weights = torch.tensor(weights, dtype=torch.float32)
        self.replacement = replacement
        self.num_replicas = num_replicas
        self.rank = rank
        self.drop_last = drop_last
        if self.drop_last and num_samples % self.num_replicas != 0:
            self.num_samples = math.ceil((num_samples - self.num_replicas) / self.num_replicas)
        else:
            self.num_samples = math.ceil(num_samples / self.num_replicas)
        self.total_size = self.num_samples * self.num_replicas
        self.shuffle = shuffle
        self.seed = seed
        # When a DataLoader with num_workers>0 and persistent_workers=True is used,
        # each persistent worker holds a copy of the sampler object. Calls of set_epoch()
        # would thus not be reflected in the worker processes. In order to still enable
        # an update of the epoch number, a synchronized value in shared memory is used.
        self._epoch = mp.Value("Q", 0)
        # We cannot sampler more indices than available if replacement=False
        if not replacement and self.total_size > len(self.weights):
            raise ValueError(
                f"{type(self).__name__}() total number of samples is greater than number of available"
                " samples to draw without replacement; reduce 'num_samples' and/or enable 'drop_last'"
            )

    @property
    def epoch(self) -> int:
        return self._epoch.value

    def set_epoch(self, epoch: int) -> None:
        r"""Set training epoch used to adjust seed of number generator when shuffling is enabled."""
        if epoch < 0:
            raise ValueError(f"{type(self).__name__}.set_epoch() 'epoch' must be non-negative")
        self._epoch.value = epoch

    def __len__(self) -> int:
        return self.num_samples

    def __iter__(self) -> Iterator[T_co]:
        g = torch.Generator()
        g = g.manual_seed(self.seed + (self.epoch if self.shuffle else 0))
        indices = torch.multinomial(
            self.weights, self.total_size, replacement=self.replacement, generator=g
        )
        indices = indices.tolist()
        if self.shuffle and not self.replacement:
            perm = torch.randperm(len(indices), generator=g).tolist()
            indices = [indices[j] for j in perm]
        assert len(indices) == self.total_size
        indices = indices[self.rank : self.total_size : self.num_replicas]
        assert len(indices) == self.num_samples
        return iter(indices)



r"""Interfaces and utilities for AWS Simple Storage Service (S3)."""

from .config import S3Config
from .client import S3Client
from .object import S3Object


__all__ = ("S3Config", "S3Client", "S3Object")



r"""Utility functions for creating argparse based command-line interfaces.

In order to implement a CLI module, first define an argument parser by implementing
a ``parser()`` function within the module which instantiates an ``ArgumentParser``
with the respective command-line arguments of the CLI module.

.. code::

    def parser(*args, **kwargs) -> ArgumentParser:
        ...

Next, implement the main function of the CLI module. To indicate that this function
receives already parsed command-line arguments, and in accordance with the respective
attribute of an argparse subparsers, this function must be named ``func()``. In addition,
an ``init()`` function can optionally be defined. It can be used, for example, to
configure the loggers and to set environment variables before ``func()`` is invoked.
These initialization steps can also be done in ``func()``, but by splitting these out
into ``init()`` it is possible to implement a CLI module which calls the ``func()`` of
other CLI modules without repeating these initialization steps. The return value of
``init()`` and ``func()`` must be the exit code of the program.

.. code::

    logger = logging.getLogger()

    def init(args: ParsedArguments) -> int:
        configure_logging(args, logger)
        init_omp_num_threads(args.threads)

    def func(args: ParsedArguments) -> int:
        ...
        return 0

The above functions are sufficient to implement a CLI module. Finally, the description
of the CLI module which is printed as part of the default help output should be given
as docstring at the top of the module.

Lastly, an actual ``main()`` function which invokes the CLI module has to be defined
and either be called explicitly when ``__name__ == "__main__"``, or in a special module
file named ``__main__.py`` which is executed by the Python interpreter when the package
containing the CLI module is being executed using ``python -m``.

The ``main_func()`` closure can be used to implement a script for a single CLI, i.e.,

.. code::

    main = main_func(parser, func, init=init)

    if __name__ == "__main__":
        sys.exit(main())

The ``entry_point()`` closure can be used to implement a script which can invoke multiple CLIs.
Which CLI is being executed depends on the subcommand name. Subcommands can further be nested.
An example of such nested subcommands is given by ``conda``, e.g., ``conda env config vars set``.
This can be implemented using the ``entry_point()`` defined by this module as follows.

Implement the CLI module ``conda/env/config/vars/set.py`` which must define the ``parser()`` and
``func()`` function as detailed above. Then copy the following code into ``conda/__main__.py``

.. code::

    import sys

    from hfresearch.cli.argparse import entry_point

    main = entry_point(
        package=__name__,
        path=__path__,
        description="...",
        subcommands=["env.config.vars.set"],
    )

    sys.exit(main())

The ``subcommands`` keyword argument is optional. If not specified, ``entry_point()`` will
auto-discover the available CLI modules when ``--help`` is requested. By specifying the
available subcommands explicitly, the help output is generated faster.

"""

import argparse
import inspect
import os
import sys
from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Union
from typing_extensions import Protocol
from collections import namedtuple
from importlib import import_module
from pkgutil import walk_packages
from pkg_resources import DistributionNotFound, get_distribution

from .psutil import memory_limit
from .typing import BoolStr


ParsedArguments = argparse.Namespace  # for type annotations
UnknownArguments = List[str]

Args = ParsedArguments


STANDARD_ARGUMENTS_GROUP_TITLE = "Standard arguments"

# Valid values for arguments of type BoolStr
TRUE = (True, "1", "yes", "on", "true", "True", "TRUE")
FALSE = (False, "0", "no", "off", "false", "False", "FALSE")


class HelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
    # See: https://stackoverflow.com/a/35848313

    def add_usage(self, usage, actions, groups, prefix=None):
        if prefix is None:
            prefix = "Usage: "
        return super().add_usage(usage, actions, groups, prefix)


class ArgumentParser(argparse.ArgumentParser):
    """Customized ArgumentParser subclass to be propagated to subparsers."""

    def __init__(self, add_help=False, formatter_class=None, version=None, **kwargs):
        if formatter_class is None:
            formatter_class = HelpFormatter
        super().__init__(add_help=False, formatter_class=formatter_class, **kwargs)
        # See: https://stackoverflow.com/a/35848313
        self._positionals.title = "Positional arguments"
        self._optionals.title = "Optional arguments"
        group = self.add_argument_group(STANDARD_ARGUMENTS_GROUP_TITLE)
        if version:
            group.add_argument(
                "--version",
                action="version",
                version="%(prog)s " + version,
                help="Show program version number and exit.",
            )
        if add_help:
            group.add_argument(
                "-h",
                "--help",
                action="help",
                default=argparse.SUPPRESS,
                help="Show this help message and exit.",
            )


class ParserCallable(Protocol):
    r"""Type annotation for argument parser constructor."""

    def __call__(self, **kwargs) -> ArgumentParser:
        ...


class MainCallable(Protocol):
    r"""Type annotation for main function."""

    def __call__(self, args: Optional[List[Any]] = None) -> int:
        ...


def entry_point(
    package: str,
    path: Iterable[str],
    description: Optional[str] = None,
    subcommands: Optional[Union[Dict[str, str], Sequence[str]]] = None,
) -> Callable[[Optional[List[Any]]], int]:
    r"""Create entry point for running named subcommand with given arguments.

    Args:
        package: Name of package containing CLI modules.
        path: Package paths containing CLI modules.
        description: Description of entry point command.
        subcommands: Description of subcommands. Dictionary keys are subcommand module
            names excluding the ``package`` name of the main CLI module. The included
            Python modules must define functions ``parser()`` and ``func()`` within
            their global scope. If ``None``, all Python subpackages and modules within
            the ``package`` are added automatically.

    Returns:
        main: Main entry point of package CLI which takes a list
            of command arguments as ``argv`` argument and returns an exit code.

    """
    if subcommands and not isinstance(subcommands, dict):
        subcommands = {name: "" for name in sorted(subcommands)}

    def main(argv: Optional[List[Any]] = None) -> int:
        r"""Run named subcommand with given arguments.

        Args:
            argv (list): Command-line arguments. If None, sys.argv[1:] is used.

        Returns:
            exit_code (int): Exit code of subcommand.

        """
        if argv is None:
            argv = sys.argv[1:]
        argv = [str(arg) for arg in argv]

        # Special 'help' subcommand or no argument
        if not argv:
            argv = ["-h"]
        elif argv[0] == "help":
            argv = argv[1:] + ["-h"]

        # CLI information
        dist = _pypi_package_name(package)
        if "__main__" in sys.argv[0]:
            prog = f"{os.path.basename(sys.executable)} -m {package}"
        else:
            prog = os.path.basename(sys.argv[0])
        try:
            version = get_distribution(dist).version
        except DistributionNotFound:
            version = None

        # Main parser
        mainparser = ArgumentParser(
            prog=prog, description=description, add_help=True, version=version
        )
        subparsers = mainparser.add_subparsers()

        default_options_parser = ArgumentParser(add_help=False)
        group = default_options_parser.add_argument_group(STANDARD_ARGUMENTS_GROUP_TITLE)
        group.add_argument(
            "-h",
            "--help",
            action="help",
            default=argparse.SUPPRESS,
            help="Show this help message and exit.",
        )
        if version:
            group.add_argument(
                "--version",
                action="version",
                version="%(prog)s " + version,
                help="Show program version number and exit.",
            )

        # Discover CLIs
        CommandInfo = namedtuple("CommandInfo", ["module", "commands"])

        def find_commands(name: str, path: List[str]) -> Dict[str, CommandInfo]:
            commands = {}
            module_infos = [
                (basename, ispkg)
                for _, basename, ispkg in walk_packages(path)
                if not basename.startswith("_")
            ]
            for basename, ispkg in module_infos:
                module_name = ".".join([name, basename])
                command_name = basename.replace("_", "-")
                commands[command_name] = CommandInfo(
                    module=module_name,
                    commands=find_commands(
                        name=module_name, path=[os.path.join(prefix, basename) for prefix in path]
                    )
                    if ispkg
                    else None,
                )
            return commands

        if subcommands:
            commands = {}
            for module_name in subcommands.keys():
                name_parts = module_name.split(".")
                command_name = name_parts[-1].replace("_", "-")
                parent_module = package
                parent_commands = commands
                for parent_name in name_parts[:-1]:
                    parent_module = ".".join([parent_module, parent_name])
                    parent_command = parent_name.replace("_", "-")
                    if parent_command not in parent_commands:
                        parent_commands[parent_command] = CommandInfo(
                            module=parent_module, commands={}
                        )
                    parent_commands = parent_commands[parent_command].commands
                parent_commands[command_name] = CommandInfo(
                    module=".".join([package, module_name]), commands=None
                )
        else:
            commands = find_commands(name=package, path=path)

        # Get command description for help output
        def get_description(module):
            description = None
            if subcommands and module.startswith(package + "."):
                description = subcommands.get(module[len(package) + 1 :])
            if description is None:
                module = import_module(module)
                description = module.__doc__
                parser_fn = getattr(module, "parser", None)
                if parser_fn is not None:
                    description = parser_fn(add_help=False).description
            return description or ""

        # Add subparsers, while avoiding import of unused CLIs
        def add_commands(subparsers, commands, argv):
            if not argv:
                return
            if argv[0] in commands:
                command_name = argv[0]
                command_info = commands[command_name]
                module = import_module(command_info.module)
                if command_info.commands is None:
                    parser_fn = getattr(module, "parser")
                    parser = parser_fn(add_help=False)
                    func = getattr(module, "func")
                    func = _func_wrapper(func=func, init=getattr(module, "init", None))
                    subparsers.add_parser(
                        command_name,
                        parents=[parser, default_options_parser],
                        help=parser.description,
                    ).set_defaults(func=func)
                else:
                    add_commands(
                        subparsers.add_parser(
                            command_name, parents=[default_options_parser], help=module.__doc__
                        ).add_subparsers(),
                        command_info.commands,
                        argv[1:],
                    )
            else:
                for command_name, command_info in commands.items():
                    subparsers.add_parser(
                        command_name,
                        parents=[default_options_parser],
                        help=get_description(command_info.module),
                    )

        add_commands(subparsers, commands, argv)

        # Parse command arguments and call requested subcommand function
        args, unknown = mainparser.parse_known_args(argv)

        if hasattr(args, "func"):
            try:
                exit_code = args.func(args, unknown)
            except KeyboardInterrupt:
                sys.stderr.write("Interrupted by user\n")
                exit_code = 1
        else:
            mainparser.print_usage()
            exit_code = 1

        return exit_code

    # Return closure
    return main


def main_func(
    parser: ParserCallable,
    func: Callable[[ParsedArguments], int],
    init: Callable[[ParsedArguments], int] = None,
) -> MainCallable:
    r"""Create main function of subcommand.

    Args:
        parser (callable): Builder function of ``ArgumentParser`` object.
        func (callable): Function accepting ``ParsedArguments`` and returning exit code.
        init (callable): Function accepting ``ParsedArguments`` and returning exit code.
            This initialization function is called before ``func`` if provided.

    Returns:
        main (callable): Main entry point of subcommand which takes a list
            of command arguments as ``argv`` argument and returns an exit code.

    """

    func = _func_wrapper(func, init=init)

    def main(argv: Optional[List[Any]] = None) -> int:
        r"""Call main function with parsed CLI arguments.

        Args:
            argv (list): Command-line arguments. If None, sys.argv[1:] is used.

        Returns:
            Exit code, zero on success. Note that not every CLI must return an exit code,
            but throw an exception in case of an error instead. If the exception is not caught,
            it will result in a non-zero exit code of the Python interpreter command.
            Use the exit code return value for different success (or error) codes if needed.

        """
        argv = None if argv is None else [str(arg) for arg in argv]
        try:
            p = parser(add_help=True)
            args = p.parse_args(argv)
            exit_code = func(args)
        except KeyboardInterrupt:
            sys.stderr.write("Interrupted by user\n")
            exit_code = 1
        return exit_code

    # Return closure
    return main


def _func_wrapper(
    func: Callable[[ParsedArguments], int], init: Callable[[ParsedArguments], int] = None
) -> Callable[[ParsedArguments], int]:
    r"""Wrap command 'init' and 'func' callables."""

    nparams = len(inspect.signature(func).parameters)
    assert nparams == 1 or nparams == 2, "func() must have one or two parameters"

    def call_init_then_func(args: ParsedArguments, unknown: UnknownArguments = []) -> int:
        exit_code = 0
        if nparams == 1 and unknown:
            raise RuntimeError(f"Unparsed arguments: {unknown}")
        if init is not None:
            exit_code = init(args)
        if exit_code == 0:
            if nparams == 2:
                exit_code = func(args, unknown)
            else:
                exit_code = func(args)
        return exit_code

    return call_init_then_func


def _pypi_package_name(module: str) -> str:
    r"""Get Python package name given CLI module __name__."""
    parts = []
    for part in module.split("."):
        if part in ("app", "apps", "cli"):
            break
        parts.append(part)
    return ".".join(parts)


def bool_from_arg(arg: BoolStr) -> bool:
    r"""Convert string argument to boolean value."""
    if arg in TRUE:
        return True
    if arg in FALSE:
        return False
    raise ValueError("bool_from_arg() 'arg' must be valid boolean string argument")


def memory_from_arg(arg: Union[int, str, None]) -> int:
    r"""Parse --memory option argument.

    Args:
        arg: Amount of memory. If None, returns the total available memory.

    Returns:
        Amount of memory in number of bytes.

    """
    if arg is None:
        return memory_limit()
    if isinstance(arg, int):
        return arg
    if not isinstance(arg, str):
        raise TypeError("memory_from_arg() 'arg' must be int or str")
    if arg.endswith("B"):
        return int(arg[:-1])
    if arg.endswith("K"):
        return int(float(arg[:-1]) * 1024)
    if arg.endswith("M"):
        return int(float(arg[:-1]) * 1024**2)
    if arg.endswith("G"):
        return int(float(arg[:-1]) * 1024**3)
    return int(arg)



r"""Modules which apply a spatial transformation to a given input data tensor.

A spatial transformer applies a :class:`.SpatialTransform`, which takes as input point coordinates
and maps these to new spatial locations, to a given input data tensor.

"""

from __future__ import annotations

from copy import copy as shallow_copy
from typing import Dict, Optional, Tuple, TypeVar, Union, overload

from torch import Tensor
from torch.nn import Module

from deepali.core.enum import PaddingMode, Sampling
from deepali.core.grid import Axes, Grid
from deepali.core.typing import Scalar
from deepali.modules import SampleImage

from .base import SpatialTransform


TSpatialTransformer = TypeVar("TSpatialTransformer", bound="SpatialTransformer")


class SpatialTransformer(Module):
    r"""Spatially transform input data.

    A :class:`.SpatialTransformer` applies a :class:`.SpatialTransform` to a given input. How the spatial
    transformation is applied to produce a transformed output is determined by the type of spatial transformer.

    The forward method of a spatial transformer invokes the spatial transform as a functor such that any registered
    forward pre- and post-hooks are executed as part of the spatial transform evaluation. This includes in particular
    the :meth:`.SpatialTransform.update` function if it is registered as a forward pre-hook. Note that this hook is by
    default installed during initialization of a spatial transform. When the update of spatial transform parameters,
    which may be inferred by a neural network, is done explicitly by the application, use :meth:`.SpatialTransform.remove_update_hook`
    to remove this forward pre-hook before subsequent evaluations of the spatial transform. When doing so, ensure to
    update the parameters when necessary using either :meth:`.SpatialTransformer.update` or :meth:`.SpatialTransform.update`.

    """

    def __init__(self, transform: SpatialTransform) -> None:
        r"""Initialize spatial transformer.

        Args:
            transform: Spatial coordinate transformation.

        """
        if not isinstance(transform, SpatialTransform):
            raise TypeError(
                f"{type(self).__name__}() requires 'transform' of type SpatialTransform"
            )
        super().__init__()
        self._transform = transform

    @property
    def transform(self) -> SpatialTransform:
        r"""Spatial grid transformation."""
        return self._transform

    @overload
    def condition(self) -> Tuple[tuple, dict]:
        r"""Get arguments on which transformation is conditioned.

        Returns:
            args: Positional arguments.
            kwargs: Keyword arguments.

        """
        ...

    @overload
    def condition(self: TSpatialTransformer, *args, **kwargs) -> TSpatialTransformer:
        r"""Get new transformation which is conditioned on the specified arguments."""
        ...

    def condition(
        self: TSpatialTransformer, *args, **kwargs
    ) -> Union[TSpatialTransformer, Tuple[tuple, dict]]:
        r"""Get or set data tensors and parameters on which transformation is conditioned."""
        if args:
            return shallow_copy(self).condition_(*args)
        return self._transform.condition()

    def condition_(self: TSpatialTransformer, *args, **kwargs) -> TSpatialTransformer:
        r"""Set data tensors and parameters on which this transformation is conditioned."""
        self._transform.condition_(*args, **kwargs)
        return self

    def update(self: TSpatialTransformer) -> TSpatialTransformer:
        r"""Update internal state of spatial transformation (cf. :meth:`.SpatialTransform.update`)."""
        self._transform.update()
        return self


class ImageTransformer(SpatialTransformer):
    r"""Spatially transform an image.

    The :class:`.ImageTransformer` applies a :class:`.SpatialTransform` to the sampling grid
    points of the target domain, optionally followed by linear transformation from target to
    source domain, and samples the input image of shape ``(N, C, ..., X)`` at these deformed
    source grid points. If the spatial transformation is non-rigid, this is also commonly
    referred to as warping the input image.

    Note that the :meth:`.ImageTransformer.forward` method invokes the spatial transform as
    functor, i.e., it triggers any pre-forward and post-forward hooks that are registered with
    the spatial transform when evaluating it. This in particular includes the forward pre-hook
    that invokes :meth:`.SpatialTransform.update` (cf. :class:`.SpatialTransformer`).

    """

    def __init__(
        self,
        transform: SpatialTransform,
        target: Optional[Grid] = None,
        source: Optional[Grid] = None,
        sampling: Union[Sampling, str] = Sampling.LINEAR,
        padding: Union[PaddingMode, str, Scalar] = PaddingMode.BORDER,
        align_centers: bool = False,
        flip_coords: bool = False,
    ) -> None:
        r"""Initialize spatial image transformer.

        Args:
            transform: Spatial coordinate transformation which is applied to ``target`` grid points.
            target: Sampling grid of output images. Use ``transform.grid()`` if ``None``.
            source: Sampling grid of input images. The attributes of this sampling grid are used
                to map the transformed normalized coordinates defined with respect to ``transform.grid()``
                to normalized coordinates with respect to the moving ``source`` image grid. These
                normalized coordinates are used to sample the image tensor passed to the forward
                method of the image transformer. If ``None``, it is assumed that the sampling grid
                of the moving source image is identical to the ``target`` image grid. In order to
                avoid having to re-normalize coordinates between image domains, the ``source`` grid
                domain should be the same as the ``transform.grid()`` (cf. :meth:`Grid.same_domain_as`).
            sampling: Image interpolation mode.
            padding: Image extrapolation mode or scalar out-of-domain value.
            align_centers: Whether to implicitly align the ``target`` and ``source`` centers.
                If ``True``, only the affine component of the target to source transformation
                is applied after the spatial grid ``transform``. If ``False``, also the
                translation of grid center points is considered.
            flip_coords: Whether spatial transformation applies to flipped grid point coordinates
                in the order (z, y, x). The default is grid point coordinates in the order (x, y, z).

        """
        super().__init__(transform)
        if target is None:
            target = transform.grid()
        if source is None:
            source = target
        if not isinstance(target, Grid):
            raise TypeError(f"{type(self).__name__}() 'target' must be of type Grid")
        if not isinstance(source, Grid):
            raise TypeError(f"{type(self).__name__}() 'source' must be of type Grid")
        device = transform.device
        sampler = SampleImage(
            target=transform.grid(),
            source=source,
            sampling=sampling,
            padding=padding,
            align_centers=align_centers,
        )
        self._sample = sampler.to(device)
        self._target_grid = target
        self._flip_coords = bool(flip_coords)
        x = target.coords(align_corners=transform.align_corners(), flip=flip_coords, device=device)
        x = target.transform_points(x, axes=transform.axes(), to_grid=transform.grid())
        self.register_buffer("grid_coords", x.unsqueeze(0), persistent=False)

    @property
    def sample(self) -> SampleImage:
        r"""Source image sampler."""
        return self._sample

    def target_grid(self) -> Grid:
        r"""Sampling grid of output images."""
        return self._target_grid

    def source_grid(self) -> Grid:
        r"""Sampling grid of input images."""
        return self._sample.source_grid()

    def align_centers(self) -> bool:
        r"""Whether grid center points are implicitly aligned."""
        return self._sample.align_centers()

    @overload
    def forward(self, data: Tensor) -> Tensor:
        r"""Sample batch of images at spatially transformed target grid points."""
        ...

    @overload
    def forward(self, data: Tensor, mask: Tensor) -> Tuple[Tensor, Tensor]:
        r"""Sample batch of masked images at spatially transformed target grid points."""
        ...

    @overload
    def forward(self, data: Dict[str, Union[Tensor, Grid]]) -> Dict[str, Union[Tensor, Grid]]:
        r"""Sample batch of optionally masked images at spatially transformed target grid points."""
        ...

    def forward(
        self,
        data: Union[Tensor, Dict[str, Union[Tensor, Grid]]],
        mask: Optional[Tensor] = None,
    ) -> Union[Tensor, Tuple[Tensor, Tensor], Dict[str, Union[Tensor, Grid]]]:
        r"""Sample batch of images at spatially transformed target grid points."""
        grid: Tensor = self.grid_coords
        grid = self._transform(grid, grid=True)
        if self._flip_coords:
            grid = grid.flip((-1,))
        return self._sample(grid, data, mask)


class PointSetTransformer(SpatialTransformer):
    r"""Spatially transform a set of points.

    The :class:`.PointSetTransformer` applies a :class:`.SpatialTransform` to a set of input points
    with coordinates defined with respect to a specified target domain. This coordinate map may
    further be followed by a linear transformation from the grid domain of the spatial transform
    to a given source domain. When no spatial transform is given, use :func:`.grid_transform_points`.

    The forward method of a point set transformer performs the same operation as :meth:`.SpatialTransform.points`,
    but with the target and source domain arguments specified during transformer initialization. In addition,
    the point set transformer module invokes the spatial transform as a functor such that any registered
    forward pre- and post-hooks are executed as part of the spatial transform evaluation. This includes the
    forward pre-hook that invokes :meth:`.SpatialTransform.update` (cf. :class:`.SpatialTransformer`).

    """

    def __init__(
        self,
        transform: SpatialTransform,
        grid: Optional[Grid] = None,
        axes: Optional[Union[Axes, str]] = None,
        to_grid: Optional[Grid] = None,
        to_axes: Optional[Union[Axes, str]] = None,
    ) -> None:
        r"""Initialize point set transformer.

        Args:
            transform: Spatial coordinate transformation which is applied to input points.
            grid: Grid with respect to which input points are defined. Uses ``transform.grid()`` if ``None``.
            axes: Coordinate axes with respect to which input points are defined. Uses ``transform.axes()`` if ``None``.
            to_grid: Grid with respect to which output points are defined. Same as ``grid`` if ``None``.
            to_axes: Coordinate axes to which input points should be mapped to. Same as ``axes`` if ``None``.

        """
        super().__init__(transform)
        if grid is None:
            grid = transform.grid()
        if axes is None:
            axes = transform.axes()
        else:
            axes = Axes.from_arg(axes)
        if to_grid is None:
            to_grid = grid
        if to_axes is None:
            to_axes = axes
        else:
            to_axes = Axes.from_arg(to_axes)
        self._grid = grid
        self._axes = axes
        self._to_grid = to_grid
        self._to_axes = to_axes

    def target_axes(self) -> Axes:
        r"""Coordinate axes with respect to which input points are defined."""
        return self._axes

    def target_grid(self) -> Grid:
        r"""Sampling grid with respect to which input points are defined."""
        return self._grid

    def source_axes(self) -> Axes:
        r"""Coordinate axes with respect to which output points are defined."""
        return self._to_axes

    def source_grid(self) -> Grid:
        r"""Sampling grid with respect to which output points are defined."""
        return self._to_grid

    def forward(self, points: Tensor) -> Tensor:
        r"""Spatially transform a set of points."""
        transform = self.transform
        points = self._grid.transform_points(
            points,
            axes=self._axes,
            to_grid=transform.grid(),
            to_axes=transform.axes(),
            decimals=None,
        )
        points = transform(points)
        points = transform.grid().transform_points(
            points,
            axes=transform.axes(),
            to_grid=self._to_grid,
            to_axes=self._to_axes,
            decimals=None,
        )
        return points



import pytest

import numpy as np
import SimpleITK as sitk

from deepali.core.affine import euler_rotation_matrix

from deepali.core.tempfile import temp_dir
from deepali.utils.imageio.meta import meta_image_bytes
from deepali.utils.imageio.meta import read_meta_image_from_fileobj


@pytest.fixture
def rand_image() -> sitk.Image:
    rng = np.random.default_rng(123456789)
    data = (255 * rng.random((32, 64, 128))).astype(np.uint8)
    im = sitk.GetImageFromArray(data)
    rot = euler_rotation_matrix((0.1, 0.6, -0.4))
    im.SetDirection(rot.flatten().tolist())
    im.SetSpacing((1.2, 0.6, 0.25))
    return im


def test_read_meta_image(rand_image: sitk.Image) -> None:
    r"""Test reading image from MetaIO file bytes."""
    image_data = sitk.GetArrayViewFromImage(rand_image)
    with temp_dir(prefix="test_read_meta_image_") as temp:
        path = temp / "rand_image.mha"
        compress = True
        sitk.WriteImage(rand_image, str(path), compress)
        with path.open("rb") as f:
            data, meta = read_meta_image_from_fileobj(f)
        assert meta.get("NDims") == 3
        assert meta.get("ObjectType") == "Image"
        assert np.all(meta.get("DimSize") == rand_image.GetSize())
        assert meta.get("ElementNumberOfChannels") == 1
        assert np.all(meta.get("ElementSpacing") == rand_image.GetSpacing())
        assert np.all(meta.get("Offset") == rand_image.GetOrigin())
        assert np.all(meta.get("TransformMatrix").flatten() == rand_image.GetDirection())
        assert np.all(meta.get("CenterOfRotation") == (0, 0, 0))
        assert meta.get("ElementType") == image_data.dtype
        assert np.allclose(image_data, data)


def test_meta_image_bytes(rand_image: sitk.Image) -> None:
    r"""Test serialization of image to MetaIO bytes."""
    image_data = sitk.GetArrayViewFromImage(rand_image)
    with temp_dir(prefix="test_meta_image_bytes_") as temp:
        path = temp / "rand_image.mha"
        compress = True
        sitk.WriteImage(rand_image, str(path), compress)
        assert path.is_file()
        with path.open("rb") as f:
            data, meta = read_meta_image_from_fileobj(f)
        path.unlink()
        assert not path.is_file()
        blob = meta_image_bytes(data, meta)
        assert path.write_bytes(blob) == len(blob)
        assert path.is_file()
        im = sitk.ReadImage(str(path))
        assert np.all(np.equal(im.GetSize(), rand_image.GetSize()))
        assert np.all(np.equal(im.GetSpacing(), rand_image.GetSpacing()))
        assert np.all(np.equal(im.GetDirection(), rand_image.GetDirection()))
        im_data = sitk.GetArrayViewFromImage(im)
        assert image_data.dtype == im_data.dtype
        assert np.allclose(image_data, im_data)



# %%
# Imports
from typing import Optional, Sequence

import matplotlib.pyplot as plt

import torch
from torch import Tensor
from torch.random import Generator

from deepali.core import Grid
import deepali.core.bspline as B
import deepali.core.functional as U


# %%
# Auxiliary functions
def random_svf(
    size: Sequence[int],
    stride: int = 1,
    generator: Optional[Generator] = None,
) -> Tensor:
    cp_grid_size = B.cubic_bspline_control_point_grid_size(size, stride=stride)
    data = torch.randn((1, 3) + cp_grid_size, generator=generator)
    data = U.fill_border(data, margin=3, value=0, inplace=True)
    return B.evaluate_cubic_bspline(data, size=size, stride=stride)


def visualize_flow(ax, flow: Tensor, label: Optional[str] = None) -> None:
    grid = Grid(shape=flow.shape[2:], align_corners=True)
    x = grid.coords(channels_last=False, dtype=u.dtype, device=u.device)
    x = U.move_dim(x.unsqueeze(0).add_(flow), 1, -1)
    target_grid = U.grid_image(shape=flow.shape[2:], inverted=True, stride=(5, 5))
    warped_grid = U.warp_image(target_grid, x)
    ax.imshow(warped_grid[0, 0, flow.shape[2] // 2], cmap="gray")
    if label:
        ax.set_title(label, fontsize=24)


# %%
# Random velocity fields
size = (128, 128, 128)
generator = torch.Generator().manual_seed(42)
v = random_svf(size, stride=8, generator=generator).mul_(0.1)


# %%
# Compute logarithm of exponentiated velocity field
bch_terms = 3
exp_steps = 5
log_steps = 5

u = U.expv(v, steps=exp_steps)
w = U.logv(u, num_iters=log_steps, bch_terms=bch_terms, exp_steps=exp_steps, sigma=1.0)

fig, axes = plt.subplots(1, 4, figsize=(40, 10))

ax = axes[0]
ax.set_title("v", fontsize=32, pad=20)
visualize_flow(ax, v)

ax = axes[1]
ax.set_title("u = exp(v)", fontsize=32, pad=20)
visualize_flow(ax, u)

ax = axes[2]
ax.set_title("log(u)", fontsize=32, pad=20)
visualize_flow(ax, w)

error = w.sub(v).norm(dim=1, keepdim=True)

ax = axes[3]
ax.set_title("|log(u) - v|", fontsize=32, pad=20)
_ = ax.imshow(error[0, 0, error.shape[2] // 2], cmap="jet", vmin=0, vmax=0.1)

print(f"Mean error:     {error.mean():.5f}")
print(f"Maximium error: {error.max():.5f}")

# %%



r"""Utility functions for `SimpleITK <https://simpleitk.org/>`_ data objects."""

from .imageio import read_image, write_image


# fmt: off
__all__ = (
    "read_image",
    "write_image",
)
# fmt: on



import itertools

import pytest

from deepali.core.enum import FlowChannelIndex, FlowDerivativeKeys
from deepali.core.enum import SpatialDerivativeKeys, SpatialDim


def test_flow_derivative_keys_from_arg() -> None:
    for D, order in itertools.product([2, 3], [0, 1, 2]):
        keys = FlowDerivativeKeys.from_arg(spatial_dims=D, order=1)
        expected = FlowDerivativeKeys.all(spatial_dims=D, order=1)
        assert keys == expected

    keys = ["du/dx", "du/dxx", "dv/dy", "dv/dxy"]
    assert FlowDerivativeKeys.from_arg(spatial_dims=2, which=keys) == keys
    assert FlowDerivativeKeys.from_arg(spatial_dims=2, which=keys, order=0) == []
    assert FlowDerivativeKeys.from_arg(spatial_dims=2, which=keys, order=1) == ["du/dx", "dv/dy"]
    assert FlowDerivativeKeys.from_arg(spatial_dims=2, which=keys, order=2) == ["du/dxx", "dv/dxy"]

    assert FlowDerivativeKeys.from_arg(spatial_dims=2, which="x") == ["du/dx", "dv/dx"]
    assert FlowDerivativeKeys.from_arg(spatial_dims=3, which="x") == ["du/dx", "dv/dx", "dw/dx"]


def test_flow_derivative_keys_split() -> None:
    assert FlowDerivativeKeys.split("du/dx") == (FlowChannelIndex.U, "x")
    assert FlowDerivativeKeys.split("du/dxy") == (FlowChannelIndex.U, "xy")
    assert FlowDerivativeKeys.split("dv/dzy") == (FlowChannelIndex.V, "zy")
    assert FlowDerivativeKeys.split("dw/dxxyz") == (FlowChannelIndex.W, "xxyz")

    assert FlowDerivativeKeys.split(["du/dxy", "dv/dzy"]) == [
        (FlowChannelIndex.U, "xy"),
        (FlowChannelIndex.V, "zy"),
    ]


def test_flow_derivative_keys_symbol() -> None:
    assert FlowDerivativeKeys.symbol(0, 0) == "du/dx"
    assert FlowDerivativeKeys.symbol(0, 1) == "du/dy"
    assert FlowDerivativeKeys.symbol(0, 2) == "du/dz"
    assert FlowDerivativeKeys.symbol(0, 3) == "du/dt"

    assert FlowDerivativeKeys.symbol(1, 0) == "dv/dx"
    assert FlowDerivativeKeys.symbol(1, 1) == "dv/dy"
    assert FlowDerivativeKeys.symbol(1, 2) == "dv/dz"
    assert FlowDerivativeKeys.symbol(1, 3) == "dv/dt"

    assert FlowDerivativeKeys.symbol(2, 0) == "dw/dx"
    assert FlowDerivativeKeys.symbol(2, 1) == "dw/dy"
    assert FlowDerivativeKeys.symbol(2, 2) == "dw/dz"
    assert FlowDerivativeKeys.symbol(2, 3) == "dw/dt"

    assert FlowDerivativeKeys.symbol("u", 0, "y", SpatialDim.Z) == "du/dxyz"
    assert FlowDerivativeKeys.symbol("v", "xyz") == "dv/dxyz"

    with pytest.raises(TypeError):
        FlowDerivativeKeys.symbol(0, 2.3)


def test_flow_derivative_keys_unique() -> None:
    value = FlowDerivativeKeys.unique("du/dx")
    assert isinstance(value, str)
    assert value == "du/dx"

    value = FlowDerivativeKeys.unique(["du/dx"])
    assert isinstance(value, set)
    assert value == {"du/dx"}

    assert FlowDerivativeKeys.unique(["du/dxy", "du/dyx"]) == {"du/dxy"}
    assert FlowDerivativeKeys.unique(["du/dz", "du/dxy", "du/dyx"]) == {"du/dxy", "du/dz"}
    assert FlowDerivativeKeys.unique(["dv/dxy", "du/dxy", "du/dyx"]) == {"du/dxy", "dv/dxy"}


def test_flow_derivative_keys_all() -> None:
    for d, order in itertools.product([2, 3], [0, 1, 2]):
        channel_keys = ["u", "v", "w"][:d]
        spatial_keys = SpatialDerivativeKeys.all(ndim=d, order=order)
        expected = [f"d{a}/d{b}" for a, b in itertools.product(channel_keys, spatial_keys)]
        assert FlowDerivativeKeys.all(spatial_dims=d, order=order) == expected


def test_flow_derivative_keys_gradient() -> None:
    assert FlowDerivativeKeys.gradient(spatial_dims=3, channel="v") == [
        "dv/dx",
        "dv/dy",
        "dv/dz",
    ]
    assert FlowDerivativeKeys.gradient(spatial_dims=3, channel=[0, "v"]) == [
        # fmt: off
        "du/dx", "du/dy", "du/dz",
        "dv/dx", "dv/dy", "dv/dz",
        # fmt: on
    ]


def test_flow_derivative_keys_jacobian() -> None:
    for d in [2, 3]:
        expected = FlowDerivativeKeys.all(spatial_dims=d, order=1)
        assert FlowDerivativeKeys.jacobian(spatial_dims=d) == expected


def test_flow_derivative_keys_divergence() -> None:
    assert FlowDerivativeKeys.divergence(spatial_dims=2) == ["du/dx", "dv/dy"]
    assert FlowDerivativeKeys.divergence(spatial_dims=3) == ["du/dx", "dv/dy", "dw/dz"]


def test_flow_derivative_keys_curvature() -> None:
    assert FlowDerivativeKeys.curvature(spatial_dims=2) == [
        # fmt: off
        "du/dxx", "du/dyy",
        "dv/dxx", "dv/dyy",
        # fmt: on
    ]
    assert FlowDerivativeKeys.curvature(spatial_dims=3) == [
        # fmt: off
        "du/dxx", "du/dyy", "du/dzz",
        "dv/dxx", "dv/dyy", "dv/dzz",
        "dw/dxx", "dw/dyy", "dw/dzz",
        # fmt: on
    ]


def test_flow_derivative_keys_hessian() -> None:
    for d, c in itertools.product([2, 3], [None, "u", "v", ["u", "v"]]):
        expected = FlowDerivativeKeys.all(spatial_dims=d, channel=c, order=2)
        assert FlowDerivativeKeys.hessian(spatial_dims=d, channel=c) == expected


def test_flow_derivative_keys_unmixed() -> None:
    for d, order in itertools.product([2, 3], [0, 1, 2]):
        if order == 0:
            expected = []
        elif order == 1:
            expected = FlowDerivativeKeys.all(spatial_dims=d, order=1)
        elif order == 2:
            expected = FlowDerivativeKeys.curvature(spatial_dims=d)
        else:
            raise AssertionError(f"unexpected order: {order}")
        assert FlowDerivativeKeys.unmixed(spatial_dims=d, order=order) == expected



# %%
import math
from timeit import default_timer as timer
from typing import Callable, Tuple

import torch
import torch.nn.functional as F
from torch import Size, Tensor

from deepali.core import functional as U
from deepali.core import bspline as B
from deepali.core.enum import SpatialDim


# %% Vector field control point coefficients
# device = torch.device("cuda:0")
device = torch.device("cpu")

in_size = Size((21,))  # (X, ...)
stride = (5,) * len(in_size)

cp_size = B.cubic_bspline_control_point_grid_size(in_size, stride)
cp_data = torch.arange(len(in_size) * cp_size.numel(), dtype=torch.float32, device=device)
cp_data = cp_data.reshape(1, len(cp_size), *tuple(reversed(cp_size)))


# %% Reference implementation based on MIRTK C++ code


def compute_bspline_indices_and_weights_1d(x: float, degree: int = 3) -> Tuple[Tensor, Tensor]:
    if degree & 1:
        i = int(math.floor(x)) - degree // 2
    else:
        i = int(math.floor(x + 0.5)) - degree // 2
    i = torch.arange(i, i + degree + 1, 1, dtype=torch.int)
    wx = torch.empty(4, dtype=torch.float)
    if degree == 3:
        w = x - i[1]
        wx[3] = (1 / 6) * w * w * w
        wx[0] = (1 / 6) + (1 / 2) * w * (w - 1) - wx[3]
        wx[2] = w + wx[0] - 2 * wx[3]
        wx[1] = 1 - wx[0] - wx[2] - wx[3]
    else:
        raise NotImplementedError(f"compute_bspline_indices_and_weights_1d() for degree={degree}")
    return i, wx


def compute_bspline_indices_and_weights_2d(
    x: float, y: float, degree: int = 3
) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
    if degree & 1:
        i = int(math.floor(x)) - degree // 2
        j = int(math.floor(y)) - degree // 2
    else:
        i = int(math.floor(x + 0.5)) - degree // 2
        j = int(math.floor(y + 0.5)) - degree // 2
    i = torch.arange(i, i + degree + 1, 1, dtype=torch.int)
    j = torch.arange(j, j + degree + 1, 1, dtype=torch.int)
    wx = torch.empty(4, dtype=torch.float)
    wy = torch.empty(4, dtype=torch.float)
    if degree == 3:
        # x
        w = x - i[1]
        wx[3] = (1 / 6) * w * w * w
        wx[0] = (1 / 6) + (1 / 2) * w * (w - 1) - wx[3]
        wx[2] = w + wx[0] - 2 * wx[3]
        wx[1] = 1 - wx[0] - wx[2] - wx[3]
        # y
        w = y - j[1]
        wy[3] = (1 / 6) * w * w * w
        wy[0] = (1 / 6) + (1 / 2) * w * (w - 1) - wy[3]
        wy[2] = w + wy[0] - 2 * wy[3]
        wy[1] = 1 - wy[0] - wy[2] - wy[3]
    else:
        raise NotImplementedError(f"compute_bspline_indices_and_weights_2d() for degree={degree}")
    return i, j, wx, wy


def interpolate_cubic_bspline_1d(data: Tensor, x: float) -> Tensor:
    degree = 3
    i, w = compute_bspline_indices_and_weights_1d(x, degree=degree)
    w = w.to(data)
    val = torch.zeros(data.shape[:2]).to(data)
    for a in range(degree + 1):
        ia: int = max(0, min(i[a].item(), data.shape[2] - 1))
        val += data[:, :, ia].mul(w[a])
    return val


def interpolate_cubic_bspline_2d(data: Tensor, x: float, y: float) -> Tensor:
    degree = 3
    i, j, wx, wy = compute_bspline_indices_and_weights_2d(x, y, degree=degree)
    wx = wx.to(data)
    wy = wy.to(data)
    val = torch.zeros(data.shape[:2]).to(data)
    for b in range(degree + 1):
        jb: int = max(0, min(j[b].item(), data.shape[2] - 1))
        for a in range(degree + 1):
            ia: int = max(0, min(i[a].item(), data.shape[3] - 1))
            val += data[..., jb, ia].mul(wx[a] * wy[b])
    return val


# %% Evaluate B-spline values at output points
D = cp_data.ndim - 2
N = cp_data.shape[0]
C = cp_data.shape[1]

conv_fn: Callable[..., Tensor] = [F.conv1d, F.conv2d, F.conv3d][D - 1]
kernels = B.bspline_interpolation_weights(
    degree=3, stride=stride, dtype=cp_data.dtype, device=cp_data.device
)

start = timer()
output = cp_data
for dim, kernel in zip((SpatialDim(dim).tensor_dim(cp_data.ndim) for dim in range(D)), kernels):
    weight = kernel.reshape((kernel.shape[0], 1, kernel.shape[1]) + (1,) * (D - 1))
    weight = weight.tile((C,) + (1,) * (weight.ndim - 1))
    output = U.move_dim(output, dim, 2)
    output = conv_fn(output, weight, groups=C)
    output = output.reshape((N, C, kernel.shape[0]) + (output.shape[2:]))
    output = output.transpose(2, 3).flatten(2, 3)
    print(output)
    output = U.move_dim(output, 2, dim)
output = output[(slice(0, N), slice(0, C)) + tuple(slice(0, n) for n in reversed(in_size))]
output = output.contiguous()
print(f"Elapsed time: {timer() - start:.3f}s")

assert output.shape[0] == N
assert output.shape[1] == C
assert output.shape[2:] == tuple(reversed(in_size))


# %%
kernel = tuple(B.cubic_bspline1d(s) for s in stride)
for _ in range(3):
    start = timer()
    result1 = B.evaluate_cubic_bspline(
        cp_data, size=in_size, stride=stride, kernel=kernel, transpose=True
    )
    print(f"Elapsed time: {timer() - start:.3f}s")

kernel = B.bspline_interpolation_weights(
    degree=3, stride=stride, dtype=cp_data.dtype, device=cp_data.device
)
for _ in range(3):
    start = timer()
    result2 = B.evaluate_cubic_bspline(
        cp_data, size=in_size, stride=stride, kernel=kernel, transpose=False
    )
    print(f"Elapsed time: {timer() - start:.3f}s")

assert torch.allclose(result1, result2, atol=0.01)


# %% Cubic B-spline kernel and its derivatives
import matplotlib.pyplot as plt

fig, ax = plt.subplots(1, 1, figsize=(12, 10))

stride = 7

cp_data = torch.zeros((1, 1, 11))
cp_data[0, 0, (cp_data.shape[2] - 1) // 2] = 1

for derivative in range(3):
    kernel = B.cubic_bspline_interpolation_weights(stride=stride, derivative=derivative)
    values = B.evaluate_cubic_bspline(cp_data, stride=stride, kernel=kernel)
    ax.plot(values[0, 0])


# %%



r"""Basic linear algebra functions, e.g., to work with homogeneous coordinate transformations."""

from enum import Enum
from functools import reduce
from operator import mul
from typing import Optional, Sequence, Tuple, Union

import torch
import torch.nn.functional as F
from torch import Tensor

from .tensor import as_tensor
from .typing import Device, DType

from ._kornia import (
    angle_axis_to_rotation_matrix,
    angle_axis_to_quaternion,
    rotation_matrix_to_angle_axis,
    rotation_matrix_to_quaternion,
    quaternion_to_angle_axis,
    quaternion_to_rotation_matrix,
    quaternion_log_to_exp,
    quaternion_exp_to_log,
    normalize_quaternion,
)


__all__ = (
    "as_homogeneous_matrix",
    "as_homogeneous_tensor",
    "hmm",
    "homogeneous_matmul",
    "homogeneous_matrix",
    "homogeneous_transform",
    "tensordot",
    "vectordot",
    "vector_rotation",
    # adapted from kornia.geometry
    "angle_axis_to_rotation_matrix",
    "angle_axis_to_quaternion",
    "rotation_matrix_to_angle_axis",
    "rotation_matrix_to_quaternion",
    "quaternion_to_angle_axis",
    "quaternion_to_rotation_matrix",
    "quaternion_log_to_exp",
    "quaternion_exp_to_log",
    "normalize_quaternion",
)


class HomogeneousTensorType(Enum):
    r"""Type of homogeneous transformation tensor."""

    AFFINE = "affine"  # Square affine transformation matrix.
    HOMOGENEOUS = "homogeneous"  # Full homogeneous transformation matrix.
    TRANSLATION = "translation"  # Translation vector.


def as_homogeneous_tensor(
    tensor: Tensor, dtype: Optional[DType] = None, device: Optional[Device] = None
) -> Tuple[Tensor, HomogeneousTensorType]:
    r"""Convert tensor to homogeneous coordinate transformation."""
    tensor_ = as_tensor(tensor, dtype=dtype, device=device)
    if tensor_.ndim == 0:
        raise ValueError("Expected at least 1-dimensional 'tensor'")
    if tensor_.ndim == 1:
        tensor_ = tensor_.unsqueeze(1)
    if tensor_.shape[-1] == 1:
        type_ = HomogeneousTensorType.TRANSLATION
    elif tensor_.shape[-1] == tensor_.shape[-2]:
        type_ = HomogeneousTensorType.AFFINE
    elif tensor_.shape[-1] == tensor_.shape[-2] + 1:
        type_ = HomogeneousTensorType.HOMOGENEOUS
    else:
        raise ValueError(f"Invalid homogeneous 'tensor' shape {tensor_.shape}")
    return tensor_, type_


def as_homogeneous_matrix(
    tensor: Tensor, dtype: Optional[DType] = None, device: Optional[Device] = None
) -> Tensor:
    r"""Convert tensor to homogeneous coordinate transformation matrix.

    Args:
        tensor: Tensor of translations of shape ``(D,)`` or ``(..., D, 1)``, tensor of square affine
            matrices of shape ``(..., D, D)``, or tensor of homogeneous transformation matrices of
            shape ``(..., D, D + 1)``.
        dtype: Data type of output matrix. If ``None``, use ``tensor.dtype`` or default.
        device: Device on which to create matrix. If ``None``, use ``tensor.device`` or default.

    Returns:
        Homogeneous coordinate transformation matrices of shape ``(..., D, D + 1)``. If ``tensor`` has already
        shape ``(..., D, D + 1)``, a reference to this tensor is returned without making a copy, unless requested
        ``dtype`` and ``device`` differ from ``tensor`` (cf. ``as_tensor()``). Use ``homogeneous_matrix()``
        if a copy of the input ``tensor`` should always be made.


    """
    tensor_, type_ = as_homogeneous_tensor(tensor, dtype=dtype, device=device)
    if type_ == HomogeneousTensorType.TRANSLATION:
        A = torch.eye(tensor_.shape[-2], dtype=tensor_.dtype, device=tensor_.device)
        tensor_ = torch.cat([A, tensor_], dim=-1)
    elif type_ == HomogeneousTensorType.AFFINE:
        t = torch.tensor(0, dtype=tensor_.dtype, device=tensor_.device).expand(
            *tensor_.shape[:-1], 1
        )
        tensor_ = torch.cat([tensor_, t], dim=-1)
    elif type_ != HomogeneousTensorType.HOMOGENEOUS:
        raise ValueError(
            "Expected 'tensor' to have shape (D,), (..., D, 1), (..., D, D) or (..., D, D + 1)"
        )
    return tensor_


def homogeneous_transform(transform: Tensor, points: Tensor, vectors: bool = False) -> Tensor:
    r"""Transform points or vectors by given homogeneous transformations.

    The data type used for matrix-vector products, as well as the data type of
    the resulting tensor, is by default set to ``points.dtype``. If ``points.dtype``
    is not a floating point data type, ``transforms.dtype`` is used instead.

    Args:
        transform: Tensor of translations of shape ``(D,)``, ``(D, 1)`` or ``(N, D, 1)``, tensor of
            affine transformation matrices of shape ``(D, D)`` or ``(N, D, D)``, or tensor of homogeneous
            matrices of shape ``(D, D + 1)`` or ``(N, D, D + 1)``, respectively. When 3-dimensional
            batch of transformation matrices is given, the size of leading dimension N must be 1
            for applying the same transformation to all points, or be equal to the leading dimension
            of ``points``, otherwise. All points within a given batch dimension are transformed by
            the matrix of matching leading index. If size of ``points`` batch dimension is one,
            the size of the leading output batch dimension is equal to the number of transforms,
            each applied to the same set of input points.
        points: Either 1-dimensional tensor of single point coordinates, or multi-dimensional tensor
            of shape ``(N, ..., D)``, where last dimension contains the spatial coordinates in the
            order ``(x, y)`` (2D) or ``(x, y, z)`` (3D), respectively.
        vectors: Whether ``points`` is tensor of vectors. If ``True``, only the affine
            component of the given ``transforms`` is applied without any translation offset.
            If ``transforms`` is a 2-dimensional tensor of translation offsets, a tensor sharing
            the data memory of the input ``points`` is returned.

    Returns:
        Tensor of transformed points/vectors with the same shape as the input ``points``, except
        for the size of the leading batch dimension if the size of the input ``points`` batch dimension
        is one, but the ``transform`` batch contains multiple transformations.

    """
    if transform.ndim == 0:
        raise TypeError("homogeneous_transform() 'transform' must be non-scalar tensor")
    if transform.ndim == 1:
        transform = transform.unsqueeze(1)
    if transform.ndim == 2:
        transform = transform.unsqueeze(0)
    N = transform.shape[0]
    D = transform.shape[1]
    if N < 1:
        raise ValueError(
            "homogeneous_transform() 'transform' size of leading dimension must not be zero"
        )
    if transform.ndim != 3 or (
        transform.shape[2] != 1 and (1 < transform.shape[2] < D or transform.shape[2] > D + 1)
    ):
        raise ValueError(
            "homogeneous_transform() 'transform' must be tensor of shape"
            + " (D,), (D, 1), (D, D), (D, D + 1) or (N, D, 1), (N, D, D), (N, D, D + 1)"
        )
    if points.ndim == 0:
        raise TypeError("'points' must be non-scalar tensor")
    if (points.ndim == 1 and len(points) != D) or (points.ndim > 1 and points.shape[-1] != D):
        raise ValueError(
            "homogeneous_transform() 'points' number of spatial dimensions does not match 'transform'"
        )
    if points.ndim == 1:
        output_shape = (N,) + points.shape if N > 1 else points.shape
        points = points.expand((N,) + points.shape)
    elif N == 1:
        output_shape = points.shape
    elif points.shape[0] == 1 or points.shape[0] == N:
        output_shape = (N,) + points.shape[1:]
        points = points.expand((N,) + points.shape[1:])
    else:
        raise ValueError(
            "homogeneous_transform() expected size of leading dimension"
            " of 'transform' and 'points' to be either 1 or equal"
        )
    points = points.reshape(N, -1, D)
    if torch.is_floating_point(points):
        transform = transform.type(points.dtype)
    else:
        points = points.type(transform.dtype)
    if transform.shape[2] == 1:
        if not vectors:
            points = points + transform[..., 0].unsqueeze(1)
    else:
        points = torch.bmm(points, transform[:, :D, :D].transpose(1, 2))
        if not vectors and transform.shape[2] == D + 1:
            points += transform[..., D].unsqueeze(1)
    return points.reshape(output_shape)


def hmm(a: Tensor, b: Tensor) -> Tensor:
    r"""Compose two homogeneous coordinate transformations.

    Args:
        a: Tensor of second homogeneous transformation.
        b: Tensor of first homogeneous transformation.

    Returns:
        Composite homogeneous transformation given by a tensor of shape ``(..., D, D + 1)``.

    See also:
        ``homogeneous_matmul()``

    """
    c = homogeneous_matmul(a, b)
    return as_homogeneous_matrix(c)


def homogeneous_matmul(*args: Tensor) -> Tensor:
    r"""Compose homogeneous coordinate transformations.

    This function performs the equivalent of a matrix-matrix product for homogeneous coordinate transformations
    given as either a translation vector (tensor of shape ``(D,)`` or ``(..., D, 1)``), a tensor of square affine
    matrices of shape ``(..., D, D)``, or a tensor of homogeneous coordinate transformation matrices of shape
    ``(..., D, D + 1)``. The size of leading dimensions must either match, or be all 1 for one of the input tensors.
    In the latter case, the same homogeneous transformation is composed with each individual trannsformation of the
    tensor with leading dimension size greater than 1.

    For example, if the shape of tensor ``a`` is either ``(D,)``, ``(D, 1)``, or ``(1, D, 1)``, and the shape of tensor
    ``b`` is ``(N, D, D)``, the translation given by ``a`` is applied after each affine transformation given by each
    matrix in grouped batch tensor ``b``, and the shape of the composite transformation tensor is ``(N, D, D + 1)``.

    Args:
        args: Tensors of homogeneous coordinate transformations, where the transformation corresponding to the first
            argument is applied last, and the transformation corresponding to the last argument is applied first.

    Returns:
        Composite homogeneous transformation given by tensor of shape ``(..., D, 1)``, ``(..., D, D)``, or  ``(..., D, D + 1)``,
        respectively, where the shape of leading dimensions is determined by input tensors.

    """
    if not args:
        raise ValueError("homogeneous_matmul() at least one argument is required")
    # Convert first input to homogeneous transformation tensor
    a = args[0]
    dtype = a.dtype
    device = a.device
    if not dtype.is_floating_point:  # type: ignore
        for b in args[1:]:
            if b.is_floating_point():
                dtype = b.dtype
                break
        if not dtype.is_floating_point:  # type: ignore
            dtype = torch.float
    a, a_type = as_homogeneous_tensor(a, dtype=dtype)
    # Successively compose transformation matrices
    D = a.shape[-2]
    for b in args[1:]:
        # Convert input to homogeneous transformation tensor
        b, b_type = as_homogeneous_tensor(b, dtype=dtype)
        if b.device != device:
            raise RuntimeError("homogeneous_matmul() tensors must be on the same 'device'")
        if b.shape[-2] != D:
            raise ValueError(
                "homogeneous_matmul() tensors have mismatching number of spatial dimensions"
                + f" ({a.shape[-2]} != {b.shape[-2]})"
            )
        # Unify shape of leading dimensions
        leading_shape = None
        a_numel = a.shape[:-2].numel()
        b_numel = b.shape[:-2].numel()
        if a_numel > 1:
            if b_numel > 1 and a.shape[:-2] != b.shape[:-2]:
                raise ValueError(
                    "Expected homogeneous tensors to have matching leading dimensions:"
                    + f" {a.shape[:-2]} != {b.shape[:-2]}"
                )
            if b.ndim > a.ndim:
                raise ValueError("Homogeneous tensors have different number of leading dimensions")
            leading_shape = a.shape[:-2]
            b = b.expand(leading_shape + b.shape[-2:])
        elif b_numel > 1:
            if a.ndim > b.ndim:
                raise ValueError("Homogeneous tensors have different number of leading dimensions")
            leading_shape = b.shape[:-2]
            a = a.expand(leading_shape + a.shape[-2:])
        elif a.ndim > b.ndim:
            leading_shape = a.shape[:-2]
            b = b.expand(a.shape[:-2] + b.shape[-2:])
        else:
            leading_shape = b.shape[:-2]
            a = a.expand(b.shape[:-2] + a.shape[-2:])
        assert leading_shape is not None
        # Compose homogeneous transformations
        a = a.reshape(-1, *a.shape[-2:])
        b = b.reshape(-1, *b.shape[-2:])
        c, c_type = None, None
        if a_type == HomogeneousTensorType.TRANSLATION:
            if b_type == HomogeneousTensorType.TRANSLATION:
                c = a + b
                c_type = HomogeneousTensorType.TRANSLATION
            elif b_type == HomogeneousTensorType.AFFINE:
                c = torch.cat([b, a], dim=-1)
                c_type = HomogeneousTensorType.HOMOGENEOUS
            elif b_type == HomogeneousTensorType.HOMOGENEOUS:
                c = b.clone()
                c[..., D] += a[..., :, 0]
                c_type = HomogeneousTensorType.HOMOGENEOUS
        elif a_type == HomogeneousTensorType.AFFINE:
            if b_type == HomogeneousTensorType.TRANSLATION:
                t = torch.bmm(a, b)
                c = torch.cat([a, t], dim=-1)
                c_type = HomogeneousTensorType.HOMOGENEOUS
            elif b_type == HomogeneousTensorType.AFFINE:
                c = torch.bmm(a, b)
                c_type = HomogeneousTensorType.AFFINE
            elif b_type == HomogeneousTensorType.HOMOGENEOUS:
                A = torch.bmm(a, b[..., :D])
                t = torch.bmm(a[..., :D], b[..., D:])
                c = torch.cat([A, t], dim=-1)
                c_type = HomogeneousTensorType.HOMOGENEOUS
        elif a_type == HomogeneousTensorType.HOMOGENEOUS:
            if b_type == HomogeneousTensorType.TRANSLATION:
                t = torch.bmm(a[..., :D], b)
                c = a.clone()
                c[..., D] += t[..., :, 0]
            elif b_type == HomogeneousTensorType.AFFINE:
                A = torch.bmm(a[..., :D], b)
                t = a[..., D:]
                c = torch.cat([A, t], dim=-1)
            elif b_type == HomogeneousTensorType.HOMOGENEOUS:
                A = torch.bmm(a[..., :D], b[..., :D])
                t = a[..., D:] + torch.bmm(a[..., :D], b[..., D:])
                c = torch.cat([A, t], dim=-1)
            c_type = HomogeneousTensorType.HOMOGENEOUS
        assert c is not None, "as_homogeneous_tensor() returned invalid 'type' enumeration value"
        assert c_type is not None
        c = c.reshape(leading_shape + c.shape[-2:])
        assert c.device == device
        a, a_type = c, c_type
    return a


def homogeneous_matrix(
    tensor: Tensor,
    offset: Optional[Tensor] = None,
    dtype: Optional[DType] = None,
    device: Optional[Device] = None,
) -> Tensor:
    r"""Convert square matrix or vector to homogeneous coordinate transformation matrix.

    Args:
        tensor: Tensor of translations of shape ``(D,)`` or ``(..., D, 1)``, tensor of square affine
            matrices of shape ``(..., D, D)``, or tensor of homogeneous transformation matrices of
            shape ``(..., D, D + 1)``.
        offset: Translation offset to add to homogeneous transformations of shape ``(..., D)``.
            If a scalar is given, this offset is used as translation along each spatial dimension.
        dtype: Data type of output matrix. If ``None``, use ``offset.dtype``.
        device: Device on which to create matrix. If ``None``, use ``offset.device``.

    Returns:
        Homogeneous coordinate transformation matrices of shape (..., D, D + 1). Always makes a copy
        of ``tensor`` even if it has already the shape of homogeneous coordinate transformation matrices.

    """
    matrix = as_homogeneous_matrix(tensor, dtype=dtype, device=device)
    if matrix is tensor:
        matrix = tensor.clone()
    if offset is not None:
        D = matrix.shape[-2]
        if offset.ndim == 0:
            offset = offset.repeat(D)
        if offset.shape[-1] != D:
            raise ValueError(
                f"Expected homogeneous_matrix() 'offset' to be scalar or have last dimension of size {D}"
            )
        matrix[..., D] += offset
    return matrix


def tensordot(
    a: Tensor,
    b: Tensor,
    dims: Union[int, Sequence[int], Tuple[Sequence[int], Sequence[int]]] = 2,
) -> Tensor:
    r"""Implements ``numpy.tensordot()`` for ``Tensor``.

    Based on https://gist.github.com/deanmark/9aec75b7dc9fa71c93c4bc85c5438777.

    """
    if isinstance(dims, int):
        axes_a = list(range(-dims, 0))
        axes_b = list(range(0, dims))
    else:
        axes_a, axes_b = dims

    if isinstance(axes_a, int):
        axes_a = [axes_a]
        na = 1
    else:
        na = len(axes_a)
        axes_a = list(axes_a)

    if isinstance(axes_b, int):
        axes_b = [axes_b]
        nb = 1
    else:
        nb = len(axes_b)
        axes_b = list(axes_b)

    a = as_tensor(a)
    b = as_tensor(b)
    as_ = a.shape
    nda = a.ndim
    bs = b.shape
    ndb = b.ndim
    equal = True
    if na != nb:
        equal = False
    else:
        for k in range(na):
            if as_[axes_a[k]] != bs[axes_b[k]]:
                equal = False
                break
            if axes_a[k] < 0:
                axes_a[k] += nda
            if axes_b[k] < 0:
                axes_b[k] += ndb
    if not equal:
        raise ValueError("shape-mismatch for sum")

    # Move the axes to sum over to the end of "a" and to the front of "b"
    notin = [k for k in range(nda) if k not in axes_a]
    newaxes_a = notin + axes_a
    N2 = 1
    for axis in axes_a:
        N2 *= as_[axis]
    newshape_a = (int(reduce(mul, [as_[ax] for ax in notin])), N2)
    olda = [as_[axis] for axis in notin]

    notin = [k for k in range(ndb) if k not in axes_b]
    newaxes_b = axes_b + notin
    N2 = 1
    for axis in axes_b:
        N2 *= bs[axis]
    newshape_b = (N2, int(reduce(mul, [bs[ax] for ax in notin])))
    oldb = [bs[axis] for axis in notin]

    at = a.permute(newaxes_a).reshape(newshape_a)
    bt = b.permute(newaxes_b).reshape(newshape_b)

    res = at.matmul(bt)
    return res.reshape(olda + oldb)


def vectordot(a: Tensor, b: Tensor, w: Optional[Tensor] = None, dim: int = -1) -> Tensor:
    r"""Inner product of vectors over specified input tensor dimension."""
    c = a.mul(b)
    if w is not None:
        c.mul(w)
    return c.sum(dim)


def vector_rotation(a: Tensor, b: Tensor) -> Tensor:
    r"""Calculate rotation matrix which aligns two 3D vectors."""
    if not isinstance(a, Tensor) or not isinstance(b, Tensor):
        raise TypeError("vector_rotation() 'a' and 'b' must be of type Tensor")
    if a.shape != b.shape:
        raise ValueError("vector_rotation() 'a' and 'b' must have identical shape")
    a = F.normalize(a, p=2, dim=-1)
    b = F.normalize(b, p=2, dim=-1)
    axis = a.cross(b, dim=-1)
    norm: Tensor = axis.norm(p=2, dim=-1, keepdim=True)
    angle_axis = axis.div(norm).mul(norm.asin())
    rotation_matrix = angle_axis_to_rotation_matrix(angle_axis)
    return rotation_matrix



r"""Basic math operations with tensors."""

from typing import Optional, Union

import torch
from torch import Tensor

from .typing import Scalar


def abspow(x: Tensor, exponent: Union[int, float]) -> Tensor:
    r"""Compute ``abs(x)**exponent``."""
    if exponent == 1:
        return x.abs()
    return x.abs().pow(exponent)


def atanh(x: Tensor) -> Tensor:
    r"""Inverse of tanh function.

    Args:
        x: Function argument.

    Returns:
        Inverse of tanh function, i.e., ``y`` where ``x = tanh(y)``.

    See also:
        https://github.com/pytorch/pytorch/issues/10324

    """
    return torch.log1p(2 * x / (1 - x)) / 2


def max_difference(source: Tensor, target: Tensor) -> Tensor:
    r"""Maximum possible intensity difference.

    Note that the two input images need not be sampled on the same grid.

    Args:
        source: Source image.
        target: Reference image.

    Returns:
        Maximum possible intensity difference.

    """
    smin, smax = source.min(), source.max()
    if target is source:
        tmin, tmax = smin, smax
    else:
        tmin, tmax = target.min(), target.max()
    return torch.max(torch.abs(smax - tmin), torch.abs(tmax - smin))


def round_decimals(tensor: Tensor, decimals: int = 0, out: Optional[Tensor] = None) -> Tensor:
    r"""Round tensor values to specified number of decimals."""
    if not decimals:
        result = torch.round(tensor, out=out)
    else:
        scale = 10**decimals
        if out is tensor:
            tensor *= scale
        else:
            tensor = tensor * scale
        result = torch.round(tensor, out=out)
        result /= scale
    return result


def threshold(data: Tensor, min: Optional[Scalar], max: Optional[Scalar] = None) -> Tensor:
    r"""Get mask for given lower and upper thresholds.

    Args:
        data: Input data tensor.
        min: Lower threshold. If ``None``, use ``data.min()``.
        max: Upper threshold. If ``None``, use ``data.max()``.

    Returns:
        Boolean tensor with same shape as ``data``, where only elements with a value
        greater than or equal ``min`` and less than or equal ``max`` are ``True``.

    """
    if min is None and max is None:
        return torch.ones_like(data, dtype=torch.bool)
    if min is None:
        return data <= max
    if max is None:
        return data >= min
    return (min <= data) & (data <= max)



r"""Functions relating to tensors representing vector fields."""

from itertools import permutations, product
from typing import Dict, Optional, Sequence, Tuple, Union

import torch
from torch import Tensor
from torch.nn import functional as F

from . import affine as A
from .enum import FlowDerivativeKeys, SpatialDerivativeKeys, PaddingMode, Sampling
from .grid import ALIGN_CORNERS, Grid
from .image import check_sample_grid, grid_reshape, grid_sample
from .image import spatial_derivatives
from .image import _image_size, zeros_image
from .itertools import is_even_permutation
from .tensor import move_dim
from .typing import Array, Device, DType, Scalar, ScalarOrTuple, Shape, Size


def affine_flow(matrix: Tensor, grid: Union[Grid, Tensor], channels_last: bool = False) -> Tensor:
    r"""Compute dense flow field from homogeneous transformation.

    Args:
        matrix: Homogeneous coordinate transformation matrices of shape ``(N, D, 1)`` (translation),
            ``(N, D, D)`` (affine), or ``(N, D, D + 1)`` (homogeneous), respectively.
        grid: Image sampling ``Grid`` or tensor of shape ``(N, ..., X, D)`` of points at
            which to sample flow fields. If an object of type ``Grid`` is given, the value
            of ``grid.align_corners()`` determines if output flow vectors are with respect to
            ``Axes.CUBE`` (False) or ``Axes.CUBE_CORNERS`` (True), respectively.
        channels_last: If ``True``, flow vector components are stored in the last dimension
            of the output tensor, and first dimension otherwise.

    Returns:
        Tensor of shape ``(N, C, ..., X)`` if ``channels_last=False`` and ``(N, ..., X, C)``, otherwise.

    """
    if matrix.ndim != 3:
        raise ValueError(
            f"affine_flow() 'matrix' must be tensor of shape (N, D, 1|D|D+1), not {matrix.shape}"
        )
    device = matrix.device
    if isinstance(grid, Grid):
        grid = grid.coords(device=device)
        grid = grid.unsqueeze(0)
    elif grid.ndim < 3:
        raise ValueError(
            f"affine_flow() 'grid' must be tensor of shape (N, ...X, D), not {grid.shape}"
        )
    assert grid.device == device
    flow = A.transform_points(matrix, grid) - grid
    if not channels_last:
        flow = move_dim(flow, -1, 1)
    assert flow.device == device
    return flow


def compose_flows(u: Tensor, v: Tensor, align_corners: bool = True) -> Tensor:
    r"""Compute composite flow field ``w = v o u = u(x) + v(x + u(x))``."""
    grid = Grid(shape=u.shape[2:], align_corners=align_corners)
    x = grid.coords(channels_last=False, dtype=u.dtype, device=u.device)
    x = move_dim(x.unsqueeze(0).add_(u), 1, -1)
    v = F.grid_sample(v, x, mode="bilinear", padding_mode="border", align_corners=align_corners)
    return u.add(v)


def compose_svfs(
    u: Tensor,
    v: Tensor,
    mode: Optional[str] = None,
    sigma: Optional[float] = None,
    spacing: Optional[Union[Scalar, Array]] = None,
    stride: Optional[ScalarOrTuple[int]] = None,
    bch_terms: int = 3,
) -> Tensor:
    r"""Approximate stationary velocity field (SVF) of composite deformation.

    The output velocity field is ``w = log(exp(v) o exp(u))``, where ``exp`` is the exponential map
    of a stationary velocity field, and ``log`` its inverse. The velocity field ``w`` is given by the
    `Baker-Campbell-Hausdorff (BCH) formula <https://en.wikipedia.org/wiki/Baker%E2%80%93Campbell%E2%80%93Hausdorff_formula>`_.

    The BCH formula with 5 Lie bracket terms (cf. ``bch_terms`` parameter) is

    .. math::

        w = v + u + \frac{1}{2} [v, u]
            + \frac{1}{12} ([v, [v, u]] - [u, [v, u]])
            + \frac{1}{48} ([[v, [v, u]], u] - [v, [u, [v, u]]])

    where

    .. math::

        [[v, [v, u]], u] - [v, [u, [v, u]]] = -2 [u, [v, [v, u]]]

    References:
    - Bossa & Olmos, 2008. A new algorithm for the computation of the group logarithm of diffeomorphisms.
        https://inria.hal.science/inria-00629873
    - Vercauteren et al., 2008. Symmetric log-domain diffeomorphic registration: A Demons-based approach.
        https://doi.org/10.1007/978-3-540-85988-8_90

    Args:
        u: First applied stationary velocity field as tensor of shape ``(N, D, ..., X)``.
        v: Second applied stationary velocity field as tensor of shape ``(N, D, ..., X)``.
        bch_terms: Number of Lie bracket terms of the BCH formula to consider.
            When 0, the returned velocity field is the sum of ``u`` and ``v``.
            This approximation is accurate if the input velocity fields commute, i.e.,
            the Lie bracket [v, u] = 0. When ``bch_terms=1``, the approximation is given by
            ``w = v + u + 1/2 [v, u]`` (note ``exp(u)`` is applied before ``exp(v)``). Formula
            ``w = v + u + \frac{1}{2} [v, u] + \frac{1}{12} ([v, [v, u]] - [u, [v, u]])`` is
            used by default, i.e., ``bch_terms=3``.
        mode: Mode of :func:`flow_derivatives()` approximation.
        sigma: Standard deviation of Gaussian used for computing spatial derivatives.
        spacing: Physical size of image voxels used to compute spatial derivatives.
        stride: Number of output grid points between control points plus one for ``mode='bspline'``.

    Returns:
        Approximation of BCH formula as tensor of shape ``(N, D, ..., X)``.

    """

    def lb(a: Tensor, b: Tensor) -> Tensor:
        return lie_bracket(a, b, mode=mode, sigma=sigma, spacing=spacing, stride=stride)

    for name, flow in [("u", u), ("v", v)]:
        if flow.ndim < 4:
            raise ValueError(
                f"compose_svfs() '{name}' must be vector field of shape (N, D, ..., X)"
            )
        if flow.shape[1] != flow.ndim - 2:
            raise ValueError(f"compose_svfs() '{name}' must have shape (N, D, ..., X)")
    if u.shape != v.shape:
        raise ValueError("compose_svfs() 'u' and 'v' must have the same shape")
    if bch_terms < 0:
        raise ValueError("compose_svfs() 'bch_terms' must not be negative")
    elif bch_terms > 5:
        raise NotImplementedError("compose_svfs() 'bch_terms' of more than 6 not implemented")

    # w = v + u
    w = v.add(u)
    if bch_terms >= 1:
        # + 1/2 [v, u]
        vu = lb(v, u)
        w = w.add(vu.mul(0.5))
    if bch_terms >= 2:
        # + 1/12 [v, [v, u]]
        vvu = lb(v, vu)
        w = w.add(vvu.mul(1 / 12))
    if bch_terms >= 3:
        # - 1/12 [u, [v, u]]
        uvu = lb(u, vu)
        w = w.sub(uvu.mul(1 / 12))
    if bch_terms >= 4:
        # + 1/48 [[v, [v, u]], u] = - 1/48 [u, [v, [v, u]]]
        # - 1/48 [v, [u, [v, u]]] = - 1/48 [u, [v, [v, u]]]
        uvvu = lb(u, vvu)
        w = w.sub(uvvu.mul((1 if bch_terms == 4 else 2) / 48))

    return w


def curl(
    flow: Tensor,
    mode: Optional[str] = None,
    sigma: Optional[float] = None,
    spacing: Optional[Union[Scalar, Array]] = None,
    stride: Optional[ScalarOrTuple[int]] = None,
) -> Tensor:
    r"""Calculate curl of vector field.

    Args:
        flow: Vector field as tensor of shape ``(N, D, ..., X)``, where ``D`` must be 3.
        mode: Mode of :func:`flow_derivatives` approximation.
        sigma: Standard deviation of Gaussian used for computing spatial derivatives.
        spacing: Physical size of image voxels used to compute spatial derivatives.
        stride: Number of output grid points between control points plus one for ``mode='bspline'``.

    Returns:
        In case of a 3D input vector field, output is another 3D vector field of rotation vectors,
        where axis of rotation corresponds to the unit vector and rotation angle to the magnitude
        of the rotation vector, as tensor of shape ``(N, 3, Z, Y, X)``.

    """
    if not isinstance(flow, Tensor):
        raise TypeError("curl() 'flow' must be of type torch.Tensor")
    if not mode:
        mode = "forward_central_backward"
    kwargs = dict(mode=mode, sigma=sigma, spacing=spacing, stride=stride)
    if flow.ndim == 4:
        if flow.shape[1] != 2:
            raise ValueError("curl() 'flow' must have shape (N, 2, Y, X)")
        which = ["du/dy", "dv/dx"]
        deriv = flow_derivatives(flow, which=which, **kwargs)
        curlv = deriv["dv/dx"].sub(deriv["du/dy"])
    elif flow.ndim == 5:
        if flow.shape[1] != 3:
            raise ValueError("curl() 'flow' must have shape (N, 3, Z, Y, X)")
        which = ["du/dy", "du/dz", "dv/dx", "dv/dz", "dw/dx", "dw/dy"]
        deriv = flow_derivatives(flow, which=which, **kwargs)
        curlv = torch.cat(
            [
                deriv["dw/dy"].sub(deriv["dv/dz"]),
                deriv["du/dz"].sub(deriv["dw/dx"]),
                deriv["dv/dx"].sub(deriv["du/dy"]),
            ],
            dim=1,
        )
    else:
        raise ValueError("curl() 'flow' must be 2- or 3-dimensional vector field")
    return curlv


def divergence(
    flow: Tensor,
    mode: Optional[str] = None,
    sigma: Optional[float] = None,
    spacing: Optional[Union[Scalar, Array]] = None,
    stride: Optional[ScalarOrTuple[int]] = None,
) -> Tensor:
    r"""Calculate divergence of vector field.

    Args:
        flow: Vector field as tensor of shape ``(N, D, ..., X)``.
        mode: Mode of :func:`flow_derivatives` approximation.
        sigma: Standard deviation of Gaussian used for computing spatial derivatives.
        spacing: Physical size of image voxels used to compute spatial derivatives.
        stride: Number of output grid points between control points plus one for ``mode='bspline'``.

    Returns:
        Scalar divergence field.

    """
    if not isinstance(flow, Tensor):
        raise TypeError("divergence() 'flow' must be of type torch.Tensor")
    if flow.ndim < 4:
        raise ValueError("divergence() 'flow' must be at least 4-dimensional tensor")
    N = flow.shape[0]
    D = flow.shape[1]
    if flow.ndim != D + 2:
        raise ValueError(
            f"divergence() 'flow' must be tensor of shape (N, {flow.ndim - 2}, ..., X)"
        )
    kwargs = dict(mode=mode, sigma=sigma, spacing=spacing, stride=stride)
    which = FlowDerivativeKeys.divergence(spatial_dims=D)
    deriv = flow_derivatives(flow, which=which, **kwargs)
    ref = deriv["du/dx"]
    div = torch.zeros((N, 1) + ref.shape[2:], dtype=ref.dtype, device=ref.device)
    for value in deriv.values():
        div = div.add_(value)
    return div


def divergence_free_flow(
    data: Tensor,
    mode: Optional[str] = None,
    sigma: Optional[float] = None,
    spacing: Optional[Union[Scalar, Array]] = None,
    stride: Optional[ScalarOrTuple[int]] = None,
) -> Tensor:
    r"""Construct divergence-free vector field from D-1 scalar fields or one 3-dimensional vector field, respectively.

    The input fields must be sufficiently smooth for the output vector field to have zero divergence. To produce a
    3-dimensional vector field, a better result may be obtained using the :func:`curl()` of another 3-dimensional
    vector field instead of two scalar fields concatenated along the channel dimension. Gaussian blurring with
    a positive ``sigma`` value or ``mode='bspline'`` may also be used to create a vector field from smoothed inputs.

    References:
        Barbarosie, Representation of divergence-free vector fields, Quart. Appl. Math. 69 (2011), 309-316
            http://dx.doi.org/10.1090/S0033-569X-2011-01215-2

    Args:
        data: A scalar field as tensor of shape ``(N, 1, Y, X)`` to generate a divergence-free 2-dimensional vector field.
            When the input is a tensor of shape ``(N, 2, Z, Y, X)``, a 3-dimensional vector field is generated using
            the cross product of the gradients of the two scalar fields. Otherwise, the input tensor must be of shape
            ``(N, 3, Z, Y, X)`` and the output is the curl of the vector field.
        mode: Mode of :func:`flow_derivatives()` approximation.
        sigma: Standard deviation of Gaussian used smooth input field.
        spacing: Physical size of image voxels used to compute finite differences.
        stride: Number of output grid points between control points plus one for ``mode='bspline'``.

    Returns:
        Divergence-free vector field as tensor of shape ``(N, D, ..., X)``.

    """
    if not isinstance(data, Tensor):
        raise TypeError("divergence_free_flow() 'data' must be of type torch.Tensor")
    if data.ndim < 4:
        raise ValueError("divergence_free_flow() 'data' must be at least 4-dimensional tensor")
    C = data.shape[1]
    D = data.ndim - 2
    if spacing is None:
        spacing = tuple(reversed([2 / (n - 1) for n in data.shape[2:]]))
    kwargs = dict(mode=mode, sigma=sigma, spacing=spacing, stride=stride)
    if D == 2 and C == 1:
        # 90 deg rotation of gradient field
        deriv = spatial_derivatives(data, order=1, **kwargs)
        flow = torch.cat([deriv["y"].neg_(), deriv["x"]], dim=1)
    elif D == 3 and C == 2:
        deriv = spatial_derivatives(data, order=1, **kwargs)
        # Following notation at https://en.wikipedia.org/wiki/Cross_product#Coordinate_notation
        a1 = deriv["x"].narrow(1, 0, 1)
        a2 = deriv["y"].narrow(1, 0, 1)
        a3 = deriv["z"].narrow(1, 0, 1)
        b1 = deriv["x"].narrow(1, 1, 1)
        b2 = deriv["y"].narrow(1, 1, 1)
        b3 = deriv["z"].narrow(1, 1, 1)
        s1 = a2.mul(b3).sub_(a3.mul(b2))
        s2 = a3.mul(b1).sub_(a1.mul(b3))
        s3 = a1.mul(b2).sub_(a2.mul(b1))
        flow = torch.cat([s1, s2, s3], dim=1)
    elif D == 3 and C == 3:
        flow = curl(data, **kwargs)
    else:
        raise ValueError(
            "divergence_free_flow() 'data' must be tensor of shape"
            " (N, 1, Y, X), (N, 2, Z, Y, X), or (N, 3, Z, Y, X)"
        )
    return flow


def expv(
    flow: Tensor,
    scale: Optional[float] = None,
    steps: Optional[int] = None,
    sampling: Union[Sampling, str] = Sampling.LINEAR,
    padding: Union[PaddingMode, str] = PaddingMode.BORDER,
    align_corners: bool = ALIGN_CORNERS,
    inverse: bool = False,
) -> Tensor:
    r"""Group exponential maps of flow fields computed using scaling and squaring.

    Args:
        flow: Batch of flow fields as tensor of shape ``(N, D, ..., X)``.
        scale: Constant flow field scaling factor.
        steps: Number of scaling and squaring steps.
        sampling: Flow field interpolation mode.
        padding: Flow field extrapolation mode.
        align_corners: Whether ``flow`` vectors are defined with respect to
            ``Axes.CUBE`` (False) or ``Axes.CUBE_CORNERS`` (True).
        inverse: Whether to negate scaled velocity field. Setting this to ``True``
            is equivalent to negating the ``scale`` (e.g., ``scale=-1``).

    Returns:
        Exponential map of input flow field. If ``steps=0``, a reference to ``flow`` is returned.

    """
    if scale is None:
        scale = 1
    if inverse:
        scale = -scale
    if steps is None:
        steps = 5
    if not isinstance(steps, int):
        raise TypeError("expv() 'steps' must be of type int")
    if steps < 0:
        raise ValueError("expv() 'steps' must be positive value")
    if steps == 0:
        return flow
    device = flow.device
    grid = Grid(shape=flow.shape[2:], align_corners=align_corners)
    grid = grid.coords(dtype=flow.dtype, device=device)
    assert grid.device == device
    disp = flow * (scale / 2**steps)
    assert disp.device == device
    for _ in range(steps):
        disp = disp + warp_image(
            disp,
            grid,
            flow=move_dim(disp, 1, -1),  # channels last
            mode=sampling,
            padding=padding,
            align_corners=align_corners,
        )
        assert disp.device == device
    return disp


def flow_derivatives(
    flow: Tensor,
    which: Optional[Union[str, Sequence[str]]] = None,
    order: Optional[int] = None,
    mode: Optional[str] = None,
    sigma: Optional[float] = None,
    spacing: Optional[Union[Scalar, Array]] = None,
    stride: Optional[ScalarOrTuple[int]] = None,
) -> Dict[str, Tensor]:
    r"""Calculate spatial derivatives of flow field.

    Args:
        flow: Flow field as tensor of shape ``(N, D, ..., X)``.
        which: String codes of spatial deriviatives to compute (cf. :class:`FlowDerivativeKeys`).
            When only a sequence of spatial dimension keys is given (cf. :class:`SpatialDerivateKeys`),
            the respective spatial derivative is computed for all vector field components, i.e.,
            "x" is shorthand for "du/dx", "dv/dx", and "dw/dx" in case of a 3-dimensional flow field.
        order: Order of spatial derivatives. When both ``which`` and ``order`` are specified,
            only the derivatives listed in ``which`` that are of the given order are returned.
        mode: Method to use for approximating of :func:`spatial_derivatives()`.
        sigma: Standard deviation of Gaussian kernel in grid units. If ``None`` or zero,
            no Gaussian smoothing is used for calculation of finite differences, and a
            default standard deviation of 0.4 is used when ``mode="gaussian"``.
        spacing: Physical spacing between image grid points, e.g., ``(sx, sy, sz)``.
            When a scalar is given, the same spacing is used for each image and spatial dimension.
            If a sequence is given, it must be of length equal to the number of spatial dimensions ``D``,
            and specify a separate spacing for each dimension in the order ``(x, ...)``. In order to
            specify a different spacing for each image in the input ``data`` batch, a 2-dimensional
            tensor must be given, where the size of the first dimension is equal to ``N``. The second
            dimension can have either size 1 for an isotropic spacing, or ``D`` in case of an
            anisotropic grid spacing.
        stride: Number of output grid points between control points plus one for ``mode='bspline'``.
            If a sequence of values is given, these must be the strides for the different spatial
            dimensions in the order ``(sx, ...)``. A stride of 1 is equivalent to evaluating partial
            derivatives only at the usually coarser resolution of the control point grid. It should
            be noted that the stride need not match the stride used to densely sample the vector
            field at a given fixed target image resolution.

    Returns:
        Mapping from partial derivative keys to spatial derivative tensors of shape ``(N, 1, ..., X)``.
        When ``mode='bspline'``, the output tensor size is reduced by three along each spatial dimension,
        and multiplied by ``stride``.

    """
    if not isinstance(flow, Tensor):
        raise TypeError("flow_derivatives() 'flow' must be torch.Tensor")
    if flow.ndim < 4:
        raise ValueError("flow_derivatives() 'flow' must be at least 4-dimensional")
    D = flow.shape[1]
    if D != flow.ndim - 2:
        raise ValueError("flow_derivatives() 'flow' must have shape (N, D, ..., X)")
    which = FlowDerivativeKeys.from_arg(spatial_dims=D, which=which, order=order)
    if spacing is None:
        spacing = tuple(reversed([2 / (n - 1) for n in flow.shape[2:]]))
    grouped_by_component = [[] for _ in range(D)]
    for channel, spatial_key in FlowDerivativeKeys.split(which):
        i = channel.index()
        if i >= D:
            raise ValueError("flow_derivatives() 'which' contains invalid spatial derivative key")
        grouped_by_component[i].append(spatial_key)
    partial_derivatives = {}
    for i, spatial_keys in enumerate(grouped_by_component):
        unique_keys = sorted(SpatialDerivativeKeys.unique(spatial_keys))
        component_derivatives = spatial_derivatives(
            flow.narrow(1, i, 1),
            which=unique_keys,
            mode=mode,
            sigma=sigma,
            spacing=spacing,
            stride=stride,
        )
        for spatial_key in spatial_keys:
            key = FlowDerivativeKeys.symbol(i, spatial_key)
            unique_key = SpatialDerivativeKeys.sorted(spatial_key)
            partial_derivatives[key] = component_derivatives[unique_key]
    return {key: partial_derivatives[key] for key in which}


def jacobian_det(
    flow: torch.Tensor,
    mode: Optional[str] = None,
    sigma: Optional[float] = None,
    spacing: Optional[Union[Scalar, Array]] = None,
    stride: Optional[ScalarOrTuple[int]] = None,
    add_identity: bool = True,
) -> Tensor:
    r"""Evaluate Jacobian determinant of spatial deformation.

    Args:
        flow: Input vector field as tensor of shape ``(N, D, ..., X)``.
        mode: Mode of :func:`flow_derivatives()` approximation.
        sigma: Standard deviation of Gaussian used for computing spatial derivatives.
        spacing: Physical size of image voxels used to compute spatial derivatives.
        stride: Number of output grid points between control points plus one for ``mode='bspline'``.
        add_identity: Whether to calculate derivatives of :math:`u(x)` (False) or the spatial
            deformation given by :math:`x + u(x)` (True), where :math:`u` is the flow field,
            by adding the identity matrix to the Jacobian of :math:`u`.

    Returns:
        Scalar field of Jacobian determinant values as tensor of shape ``(N, 1, ..., X)``.

    """
    if flow.ndim < 4:
        raise ValueError("jacobian_det() 'flow' must be dense vector field of shape (N, D, ..., X)")
    if flow.shape[1] != flow.ndim - 2:
        raise ValueError("jacobian_det() 'flow' must have shape (N, D, ..., X)")
    D = flow.shape[1]
    kwargs = dict(mode=mode, sigma=sigma, spacing=spacing, stride=stride)
    which = FlowDerivativeKeys.jacobian(spatial_dims=D)
    deriv = flow_derivatives(flow, which=which, **kwargs)
    # Add 1 to diagonal elements of Jacobian matrix, because T(x) = x + u(x)
    if add_identity:
        for i in range(D):
            deriv[FlowDerivativeKeys.symbol(i, i)].add_(1)
    if D == 2:
        a = deriv["du/dx"]
        b = deriv["du/dy"]
        c = deriv["dv/dx"]
        d = deriv["dv/dy"]
        jac = a.mul(d).sub_(b.mul(c))
    elif D == 3:
        a = deriv["du/dx"]
        b = deriv["du/dy"]
        c = deriv["du/dz"]
        d = deriv["dv/dx"]
        e = deriv["dv/dy"]
        f = deriv["dv/dz"]
        g = deriv["dw/dx"]
        h = deriv["dw/dy"]
        i = deriv["dw/dz"]
        term_1 = a.mul(e.mul(i).sub_(f.mul(h)))
        term_2 = b.mul(d.mul(i).sub_(g.mul(f)))
        term_3 = c.mul(d.mul(h).sub_(e.mul(g)))
        jac = term_1.sub(term_2).add(term_3)
    else:
        jac: Optional[Tensor] = None
        for perm in permutations(range(D)):
            term: Optional[Tensor] = None
            for i, j in zip(range(D), perm):
                dij = deriv[FlowDerivativeKeys.symbol(i, j)]
                term = dij if term is None else term.mul_(dij)
            assert term is not None
            if jac is None:
                jac = term
            elif is_even_permutation(perm):
                jac = jac.add_(term)
            else:
                jac = jac.sub_(term)
        assert jac is not None
    return jac


def jacobian_dict(
    flow: torch.Tensor,
    mode: Optional[str] = None,
    sigma: Optional[float] = None,
    spacing: Optional[Union[Scalar, Array]] = None,
    stride: Optional[ScalarOrTuple[int]] = None,
    add_identity: bool = False,
) -> Dict[Tuple[int, int], Tensor]:
    r"""Evaluate Jacobian of spatial deformation.

    Args:
        flow: Input vector field as tensor of shape ``(N, D, ..., X)``.
        mode: Mode of :func:`flow_derivatives()` approximation.
        sigma: Standard deviation of Gaussian used for computing spatial derivatives.
        spacing: Physical size of image voxels used to compute spatial derivatives.
        stride: Number of output grid points between control points plus one for ``mode='bspline'``.
        add_identity: Whether to calculate derivatives of :math:`u(x)` (False) or the spatial
            deformation given by :math:`x + u(x)` (True), where :math:`u` is the flow field,
            by adding the identity matrix to the Jacobian of :math:`u`.

    Returns:
        Dictionary of spatial derivatives with keys corresponding to (row, col) indices.

    """
    if flow.ndim < 4:
        raise ValueError("jacobian_det() 'flow' must be dense vector field of shape (N, D, ..., X)")
    if flow.shape[1] != flow.ndim - 2:
        raise ValueError("jacobian_det() 'flow' must have shape (N, D, ..., X)")
    D = flow.shape[1]
    kwargs = dict(mode=mode, sigma=sigma, spacing=spacing, stride=stride)
    which = FlowDerivativeKeys.jacobian(spatial_dims=D)
    deriv = flow_derivatives(flow, which=which, **kwargs)
    # Optionally, add 1 to diagonal elements of Jacobian matrix, because T(x) = x + u(x).
    if add_identity:
        for i in range(D):
            deriv[FlowDerivativeKeys.symbol(i, i)].add_(1)
    jac = {}
    for i, j in product(range(D), repeat=2):
        jac[(i, j)] = deriv[FlowDerivativeKeys.symbol(i, j)]
    return jac


def jacobian_matrix(
    flow: torch.Tensor,
    mode: Optional[str] = None,
    sigma: Optional[float] = None,
    spacing: Optional[Union[Scalar, Array]] = None,
    stride: Optional[ScalarOrTuple[int]] = None,
    add_identity: bool = False,
) -> Tensor:
    r"""Evaluate Jacobian of spatial deformation.

    Args:
        flow: Input vector field as tensor of shape ``(N, D, ..., X)``.
        mode: Mode of :func:`flow_derivatives()` approximation.
        sigma: Standard deviation of Gaussian used for computing spatial derivatives.
        spacing: Physical size of image voxels used to compute spatial derivatives.
        stride: Number of output grid points between control points plus one for ``mode='bspline'``.
        add_identity: Whether to calculate derivatives of :math:`u(x)` (False) or the spatial
            deformation given by :math:`x + u(x)` (True), where :math:`u` is the flow field,
            by adding the identity matrix to the Jacobian of :math:`u`.

    Returns:
        Full Jacobian matrices as tensor of shape ``(N, ..., X, D, D)``.

    """
    D = flow.ndim - 2
    deriv = jacobian_dict(
        flow,
        mode=mode,
        sigma=sigma,
        spacing=spacing,
        stride=stride,
        add_identity=add_identity,
    )
    jac = torch.cat(list(deriv.values()), dim=1)
    jac = move_dim(jac, 1, -1)
    jac = jac.reshape(jac.shape[:-1] + (D, D))
    return jac.contiguous()


def lie_bracket(
    v: Tensor,
    u: Tensor,
    mode: Optional[str] = None,
    sigma: Optional[float] = None,
    spacing: Optional[Union[Scalar, Array]] = None,
    stride: Optional[ScalarOrTuple[int]] = None,
) -> Tensor:
    r"""Lie bracket of two vector fields.

    Evaluate Lie bracket given by ``[v, u] = Jac(v) * u - Jac(u) * v`` as defined in Eq (6)
    of Vercauteren et al. (2008).

        Most authors define the Lie bracket as the opposite of (6). Numerical simulations,
        and personal communication with M. Bossa, showed the relevance of this definition.
        Future research will aim at fully understanding the reason of this discrepancy.

    References:
    - Vercauteren, 2008. Symmetric Log-Domain Diffeomorphic Registration: A Demons-based Approach.
        doi:10.1007/978-3-540-85988-8_90

    Args:
        u: Left vector field as tensor of shape ``(N, D, ..., X)``.
        v: Right vector field as tensor of shape ``(N, D, ..., X)``.
        mode: Mode of :func:`flow_derivatives()` approximation.
        sigma: Standard deviation of Gaussian used for computing spatial derivatives.
        spacing: Physical size of image voxels used to compute spatial derivatives.
        stride: Number of output grid points between control points plus one for ``mode='bspline'``.

    Returns:
        Lie bracket of vector fields as tensor of shape ``(N, D, ..., X)``.

    """
    for name, flow in [("u", u), ("v", v)]:
        if flow.ndim < 4:
            raise ValueError(f"lie_bracket() '{name}' must be vector field of shape (N, D, ..., X)")
        if flow.shape[1] != flow.ndim - 2:
            raise ValueError(f"lie_bracket() '{name}' must have shape (N, D, ..., X)")
    if u.shape != v.shape:
        raise ValueError("lie_bracket() 'u' and 'v' must have the same shape")
    jac_u = jacobian_dict(
        u,
        mode=mode,
        sigma=sigma,
        spacing=spacing,
        stride=stride,
    )
    jac_v = jacobian_dict(
        v,
        mode=mode,
        sigma=sigma,
        spacing=spacing,
        stride=stride,
    )
    D = flow.ndim - 2
    w = torch.zeros_like(u)
    for i in range(D):
        w_i = w.narrow(1, i, 1)
        for j in range(D):
            w_i = w_i.add_(jac_v[(i, j)].mul(u.narrow(1, j, 1)))
        for j in range(D):
            w_i = w_i.sub_(jac_u[(i, j)].mul(v.narrow(1, j, 1)))
    return w


def normalize_flow(
    data: Tensor,
    size: Optional[Union[Tensor, torch.Size]] = None,
    side_length: float = 2,
    align_corners: bool = ALIGN_CORNERS,
    channels_last: bool = False,
) -> Tensor:
    r"""Map vectors with respect to unnormalized grid to vectors with respect to normalized grid."""
    if not isinstance(data, Tensor):
        raise TypeError("normalize_flow() 'data' must be tensor")
    if not data.is_floating_point():
        data = data.float()
    if size is None:
        if data.ndim < 4 or data.shape[1] != data.ndim - 2:
            raise ValueError(
                "normalize_flow() 'data' must have shape (N, D, ..., X) when 'size' not given"
            )
        size = torch.Size(reversed(data.shape[2:]))  # X,...
    zero = torch.tensor(0, dtype=data.dtype, device=data.device)
    size = torch.as_tensor(size, dtype=data.dtype, device=data.device)
    size_ = size.sub(1) if align_corners else size
    if not channels_last:
        data = move_dim(data, 1, -1)
    if side_length != 1:
        data = data.mul(side_length)
    data = torch.where(size > 1, data.div(size_), zero)
    if not channels_last:
        data = move_dim(data, -1, 1)
    return data


def logv(
    flow: Tensor,
    num_iters: int = 5,
    bch_terms: int = 1,
    sigma: Optional[float] = 1.0,
    spacing: Optional[Union[Scalar, Array]] = None,
    exp_steps: Optional[int] = None,
    sampling: Union[Sampling, str] = Sampling.LINEAR,
    padding: Union[PaddingMode, str] = PaddingMode.BORDER,
    align_corners: bool = ALIGN_CORNERS,
) -> Tensor:
    r"""Group logarithmic maps of flow fields computed using algorithm by Bossa & Olsom (2008).

    References:
    - Bossa & Olmos, 2008. A new algorithm for the computation of the group logarithm of diffeomorphisms.
        https://inria.hal.science/inria-00629873

    Args:
        num_iters: Number of iterations.
        bch_terms: Number of Lie bracket terms of the Baker-Campbell-Hausdorff (BCH) formula to use
            when computing the composite of current velocity field with the correction field.
        sigma: Standard deviation of Gaussian kernel used as low-pass filter when computing spatial
            derivatives required for evaluation of Lie brackets during application of BCH formula.
        spacing: Physical size of image voxels used to compute spatial derivatives.
        exp_steps: Number of exponentiation steps to evaluate current inverse displacement field.
        sampling: Flow field interpolation mode when computing inverse displacement field.
        padding: Flow field extrapolation mode when computing inverse displacement field.
        align_corners: Whether ``flow`` vectors are defined with respect to
            ``Axes.CUBE`` (False) or ``Axes.CUBE_CORNERS`` (True).

    Returns:
        Approximate stationary velocity field which when exponentiated (cf. :func:`expv`) results
        in the given input ``flow`` field.

    """
    v = flow
    for _ in range(num_iters):
        u = expv(
            v,
            steps=exp_steps,
            sampling=sampling,
            padding=padding,
            align_corners=align_corners,
            inverse=True,
        )
        u = compose_flows(flow, u)
        v = compose_svfs(u, v, bch_terms=bch_terms, sigma=sigma, spacing=spacing)
    return v


def denormalize_flow(
    data: Tensor,
    size: Optional[Union[Tensor, torch.Size]] = None,
    side_length: float = 2,
    align_corners: bool = ALIGN_CORNERS,
    channels_last: bool = False,
) -> Tensor:
    r"""Map vectors with respect to normalized grid to vectors with respect to unnormalized grid."""
    if not isinstance(data, Tensor):
        raise TypeError("denormalize_flow() 'data' must be tensors")
    if size is None:
        if data.ndim < 4 or data.shape[1] != data.ndim - 2:
            raise ValueError(
                "denormalize_flow() 'data' must have shape (N, D, ..., X) when 'size' not given"
            )
        size = torch.Size(reversed(data.shape[2:]))  # X,...
    zero = torch.tensor(0, dtype=data.dtype, device=data.device)
    size = torch.as_tensor(size, dtype=data.dtype, device=data.device)
    size_ = size.sub(1) if align_corners else size
    if not channels_last:
        data = move_dim(data, 1, -1)
    data = torch.where(size > 1, data.mul(size_), zero)
    if side_length != 1:
        data = data.div(side_length)
    if not channels_last:
        data = move_dim(data, -1, 1)
    return data


def sample_flow(
    flow: Tensor,
    coords: Tensor,
    padding: Optional[Union[PaddingMode, str, Scalar]] = None,
    align_corners: bool = ALIGN_CORNERS,
) -> Tensor:
    r"""Sample non-rigid flow fields at given points.

    This function samples a vector field at spatial points. The ``coords`` tensor can be of any shape,
    including ``(N, M, D)``, i.e., a batch of N point sets with cardianality M. It can also be applied to
    a tensor of grid points of shape ``(N, ..., X, D)`` regardless if the grid points are located at the
    undeformed grid positions or an already deformed grid. The given non-rigid flow field is interpolated
    at the input points ``x`` using linear interpolation. These flow vectors ``u(x)`` are returned.

    Args:
        flow: Flow fields of non-rigid transformations given as tensor of shape ``(N, D, ..., X)``
            or ``(1, D, ..., X)``. If batch size is one, but the batch size of ``coords`` is greater
            than one, this single flow fields is sampled at the different sets of points.
        coords: Normalized coordinates of points given as tensor of shape ``(N, ..., D)``
            or ``(1, ..., D)``. If batch size is one, all flow fields are sampled at the same points.
        align_corners: Whether point coordinates are with respect to ``Axes.CUBE`` (False)
            or ``Axes.CUBE_CORNERS`` (True). This option is in particular passed on to the
            ``grid_sample()`` function used to sample the flow vectors at the input points.

    Returns:
        Tensor of shape ``(N, ..., D)``.

    """
    if not isinstance(flow, Tensor):
        raise TypeError("sample_flow() 'flow' must be of type torch.Tensor")
    if flow.ndim < 4:
        raise ValueError("sample_flow() 'flow' must be at least 4-dimensional tensor")
    if not isinstance(coords, Tensor):
        raise TypeError("sample_flow() 'coords' must be of type torch.Tensor")
    if coords.ndim < 2:
        raise ValueError("sample_flow() 'coords' must be at least 2-dimensional tensor")
    G = flow.shape[0]
    N = coords.shape[0] if G == 1 else G
    D = flow.shape[1]
    if coords.shape[0] not in (1, N):
        raise ValueError(f"sample_flow() 'coords' must be batch of length 1 or {N}")
    if coords.shape[-1] != D:
        raise ValueError(f"sample_flow() 'coords' must be tensor of {D}-dimensional points")
    if padding is None:
        padding = PaddingMode.BORDER
    x = coords.expand((N,) + coords.shape[1:])
    t = flow.expand((N,) + flow.shape[1:])
    g = x.reshape((N,) + (1,) * (t.ndim - 3) + (-1, D))
    u = grid_sample(t, g, padding=padding, align_corners=align_corners)
    u = move_dim(u, 1, -1)
    u = u.reshape(x.shape)
    return u


def warp_grid(flow: Tensor, grid: Tensor, align_corners: bool = ALIGN_CORNERS) -> Tensor:
    r"""Transform undeformed grid by a tensor of non-rigid flow fields.

    This function applies a non-rigid transformation to map a tensor of undeformed grid points to a
    tensor of deformed grid points with the same shape as the input tensor. The input points must be
    the positions of undeformed spatial grid points, because this function uses interpolation to
    resize the vector fields to the size of the input ``grid``. This assumes that input points ``x``
    are the coordinates of points located on a regularly spaced undeformed grid which is aligned with
    the borders of the grid domain on which the vector fields of the non-rigid transformations are
    sampled, i.e., ``y = x + u``.

    If in doubt whether the input points will be sampled regularly at grid points of the domain of
    the spatial transformation, use ``warp_points()`` instead.

    Args:
        flow: Flow fields of non-rigid transformations given as tensor of shape ``(N, D, ..., X)``
            or ``(1, D, ..., X)``. If batch size is one, but the batch size of ``points`` is greater
            than one, all point sets are transformed by the same non-rigid transformation.
        grid: Coordinates of points given as tensor of shape ``(N, ..., D)`` or ``(1, ..., D)``.
            If batch size is one, but multiple flow fields are given, this single point set is
            transformed by each non-rigid transformation to produce ``N`` output point sets.
        align_corners: Whether grid points and flow vectors are with respect to ``Axes.CUBE``
            (False) or ``Axes.CUBE_CORNERS`` (True). This option is in particular passed on to
            the ``grid_reshape()`` function used to resize the flow fields to the ``grid`` shape.

    Returns:
        Tensor of shape ``(N, ..., D)`` with coordinates of spatially transformed points.

    """
    if not isinstance(flow, Tensor):
        raise TypeError("warp_grid() 'flow' must be of type torch.Tensor")
    if flow.ndim < 4:
        raise ValueError("warp_grid() 'flow' must be at least 4-dimensional tensor")
    if not isinstance(grid, Tensor):
        raise TypeError("warp_grid() 'grid' must be of type torch.Tensor")
    if grid.ndim < 4:
        raise ValueError("warp_grid() 'grid' must be at least 4-dimensional tensor")
    G = flow.shape[0]
    N = grid.shape[0] if G == 1 else G
    D = flow.shape[1]
    if grid.shape[0] not in (1, N):
        raise ValueError(f"warp_grid() 'grid' must be batch of length 1 or {N}")
    if grid.shape[-1] != D:
        raise ValueError(f"warp_grid() 'grid' must be tensor of {D}-dimensional points")
    x = grid.expand((N,) + grid.shape[1:])
    t = flow.expand((N,) + flow.shape[1:])
    u = grid_reshape(t, grid.shape[1:-1], align_corners=align_corners)
    u = move_dim(u, 1, -1)
    y = x + u
    return y


def warp_points(flow: Tensor, coords: Tensor, align_corners: bool = ALIGN_CORNERS) -> Tensor:
    r"""Transform set of points by a tensor of non-rigid flow fields.

    This function applies a non-rigid transformation to map a tensor of spatial points to another tensor
    of spatial points of the same shape as the input tensor. Unlike ``warp_grid()``, it can be used
    to spatially transform any set of points which are defined with respect to the grid domain of the
    non-rigid transformation, including a tensor of shape ``(N, M, D)``, i.e., a batch of N point sets with
    cardianality M. It can also be applied to a tensor of grid points of shape ``(N, ..., X, D)`` regardless
    if the grid points are located at the undeformed grid positions or an already deformed grid. The given
    non-rigid flow field is interpolated at the input points ``x`` using linear interpolation. These flow
    vectors ``u(x)`` are then added to the input points, i.e., ``y = x + u(x)``.

    Args:
        flow: Flow fields of non-rigid transformations given as tensor of shape ``(N, D, ..., X)``
            or ``(1, D, ..., X)``. If batch size is one, but the batch size of ``points`` is greater
            than one, all point sets are transformed by the same non-rigid transformation.
        coords: Normalized coordinates of points given as tensor of shape ``(N, ..., D)``
            or ``(1, ..., D)``. If batch size is one, this single point set is deformed by each
            flow field to produce ``N`` output point sets.
        align_corners: Whether points and flow vectors are with respect to ``Axes.CUBE`` (False)
            or ``Axes.CUBE_CORNERS`` (True). This option is in particular passed on to the
            ``grid_sample()`` function used to sample the flow vectors at the input points.

    Returns:
        Tensor of shape ``(N, ..., D)`` with coordinates of spatially transformed points.

    """
    x = coords
    u = sample_flow(flow, coords, align_corners=align_corners)
    y = x + u
    return y


def warp_image(
    data: Tensor,
    grid: Tensor,
    flow: Optional[Tensor] = None,
    mode: Optional[Union[Sampling, str]] = None,
    padding: Optional[Union[PaddingMode, str, Scalar]] = None,
    align_corners: bool = ALIGN_CORNERS,
) -> Tensor:
    r"""Sample data at optionally displaced grid points.

    Args:
        data: Image batch tensor of shape ``(1, C, ..., X)`` or ``(N, C, ..., X)``.
        grid: Grid points tensor of shape  ``(..., X, D)``, ``(1, ..., X, D)``, or``(N, ..., X, D)``.
            Coordinates of points at which to sample ``data`` must be with respect to ``Axes.CUBE``.
        flow: Batch of flow fields of shape  ``(..., X, D)``, ``(1, ..., X, D)``, or``(N, ..., X, D)``.
            If specified, the flow field(s) are added to ``grid`` in order to displace the grid points.
        mode: Image interpolate mode.
        padding: Image extrapolation mode or constant by which to pad input ``data``.
        align_corners: Whether ``grid`` extrema ``(-1, 1)`` refer to the grid boundary
            edges (``align_corners=False``) or corner points (``align_corners=True``).

    Returns:
        Image batch tensor of sampled data with shape determined by ``grid``.

    """
    if data.ndim < 4:
        raise ValueError("warp_image() expected tensor 'data' of shape (N, C, ..., X)")
    grid = check_sample_grid("warp", data, grid)
    N = grid.shape[0]
    D = grid.shape[-1]
    if flow is not None:
        if flow.ndim == data.ndim - 1:
            flow = flow.unsqueeze(0)
        elif flow.ndim != data.ndim:
            raise ValueError(
                f"warp_image() expected 'flow' tensor with {data.ndim - 1} or {data.ndim} dimensions"
            )
        if flow.shape[0] != N:
            flow = flow.expand(N, *flow.shape[1:])
        if flow.shape[0] != N or flow.shape[-1] != D:
            msg = f"warp_image() expected tensor 'flow' of shape (..., X, {D})"
            msg += f" or (1, ..., X, {D})" if N == 1 else f" or (1|{N}, ..., X, {D})"
            raise ValueError(msg)
        grid = grid + flow
    assert data.device == grid.device
    return grid_sample(data, grid, mode=mode, padding=padding, align_corners=align_corners)


def zeros_flow(
    size: Optional[Union[int, Size, Grid]] = None,
    shape: Optional[Shape] = None,
    num: Optional[int] = None,
    dtype: Optional[DType] = None,
    device: Optional[Device] = None,
) -> Tensor:
    r"""Create batch of flow fields filled with zeros for given image batch size or grid."""
    size = _image_size("zeros_flow", size, shape)
    return zeros_image(size, num=num, channels=len(size), dtype=dtype, device=device)



import re
from typing import Union

from torch import Tensor

from .typing import TensorCollection


__all__ = (
    "TensorCollection",
    "RE_OUTPUT_KEY_INDEX",
    "re_output_key_index",
    "tensor_collection_entry",
    "get_tensor",
)


RE_OUTPUT_KEY_INDEX = r"\[([0-9]+)\]"
re_output_key_index = re.compile(RE_OUTPUT_KEY_INDEX)


def tensor_collection_entry(output: TensorCollection, key: str) -> Union[TensorCollection, Tensor]:
    r"""Get specified output entry."""
    key = re_output_key_index.sub(r".\1", key)
    for index in key.split("."):
        if isinstance(output, (list, tuple)):
            try:
                index = int(index)
            except TypeError:
                raise KeyError(f"invalid output key {key}")
        elif not index or not isinstance(output, dict):
            raise KeyError(f"invalid output key {key}")
        output = output[index]
    return output


def get_tensor(output: TensorCollection, key: str) -> Tensor:
    r"""Get tensor at specified output entry."""
    item = tensor_collection_entry(output, key)
    if not isinstance(item, Tensor):
        raise TypeError(f"get_output_tensor() entry {key} must be Tensor")
    return item



r"""Generic wrappers and mix-ins for data transforms."""

from collections.abc import KeysView
from copy import copy as shallowcopy, deepcopy
from dataclasses import is_dataclass, fields
from typing import Any, Callable, Iterable, Mapping, Optional, Union

from torch.nn import Module

from deepali.core.collections import re_output_key_index
from deepali.core.typing import is_namedtuple


__all__ = ("ItemTransform", "ItemwiseTransform")


class ItemTransform(Module):
    r"""Transform only specified item/field of dict, named tuple, tuple, list, or dataclass."""

    def __init__(
        self,
        transform: Callable,
        key: Optional[Union[int, str, KeysView, Iterable[Union[int, str]]]] = None,
        copy: bool = False,
        ignore_meta: bool = True,
        ignore_missing: bool = False,
    ):
        r"""Initialize item transformation.

        Args:
            transform: Item value transformation.
            key: Index, key, or field name of item to transform. If ``None``, empty string, or 'all',
                apply ``transform`` to all items in the input ``data`` object. Can be a nested
                key with dot ('.') character as subkey delimiters, and contain indices to access
                list or tuple items. For example, "a.b.c", "a.b.2", "a.b.c[1]", "a[1][0]",
                "a[0].c", and "a.0.c", are all valid keys as long as the data to transform has
                the appropriate entries and (nested) container types.
            copy: Whether to copy all input items. By default, a shallow copy of the input ``data``
                is made and only the item specified by ``key`` is replaced by its transformed value.
                If ``True``, a deep copy of all input ``data`` items is made.
            ignore_missing: Whether to skip processing of items with value ``None``.
            ignore_meta: Preserve 'meta' dictionary key value as is (cf. ``MetaDataset``).

        """
        super().__init__()
        self.transform = transform
        if not isinstance(key, str) and isinstance(key, (KeysView, Iterable)):
            key = set(key)
        self.key = key
        self.copy = copy
        self.ignore_meta = ignore_meta
        self.ignore_missing = ignore_missing

    def forward(self, data: Any) -> Any:
        r"""Apply transformation.

        Args:
            data: Input value, dict, tuple, list, or dataclass.

        Returns:
            Copy of ``data`` with value of ``self.key`` replaced by its transformed value.
            If ``self.key is None``, all values in the input ``data`` are transformed.
            By default, a shallow copy of ``data`` is made unless ``self.copy == True``.

        """
        if self.key in (None, "", "all", "ALL"):
            return self._apply_all(data)
        if isinstance(self.key, set):
            keys = self.key
        elif not isinstance(self.key, str) and isinstance(self.key, (KeysView, Iterable)):
            keys = set(self.key)
        else:
            keys = [self.key]
        for key in keys:
            if isinstance(key, int):
                data = self._apply_index(data, key)
            elif isinstance(key, str):
                key = re_output_key_index.sub(r".\1", key)
                data = self._apply_key(data, key)
            else:
                raise TypeError(f"{type(self).__name__}() 'key' must be None, int, or str")
        return data

    def _apply_all(self, data: Any) -> Any:
        r"""Transform all leaf items."""
        if is_dataclass(data):
            output = shallowcopy(data)
            for field in fields(data):
                value = getattr(data, field.name)
                if not self.ignore_meta or field.name != "meta":
                    value = self._apply_all(value)
                setattr(output, field.name, value)
        elif isinstance(data, Mapping):
            output = {}
            for k, v in data.items():
                if not self.ignore_meta or k != "meta":
                    v = self._apply_all(v)
                output[k] = v
        elif isinstance(data, tuple):
            output = tuple(self._apply_all(d) for d in data)
            if is_namedtuple(data):
                output = type(data)(*output)
        elif isinstance(data, list):
            output = list(self._apply_all(d) for d in data)
        elif data is None:
            if self.ignore_missing:
                output = None
            else:
                raise ValueError(f"{type(self).__name__}() value is None (ignore_missing=False)")
        else:
            output = self.transform(data)
        return output

    def _apply_index(self, data: Any, index: int) -> Any:
        r"""Transform item at specified index."""
        if not isinstance(data, (list, tuple)):
            raise TypeError(f"{type(self).__name__}() 'data' must be list or tuple when key is int")
        try:
            item = data[index]
        except IndexError:
            raise IndexError(
                f"{type(self).__name__}() 'data' sequence must have item at index {index}"
            )
        item = self.transform(item)
        args = (item if i == index else self._maybe_copy(v) for i, v in enumerate(data))
        if is_namedtuple(data):
            return type(data)(*args)
        return type(data)(args)

    def _apply_key(self, data: Any, key: str, prefix: str = "") -> Any:
        r"""Transform specified item in data map."""
        parts = key.split(".", 1)
        if len(parts) == 1:
            parts = [parts[0], ""]
        key, subkey = parts
        if not key:
            raise KeyError(
                f"{type(self).__name__}() 'key' must not be empty"
                f" (prefix={prefix!r}, key={key!r}, subkey={subkey!r})"
            )
        # Get item at specified entry
        index = None
        item = None
        if is_dataclass(data) or is_namedtuple(data):
            try:
                item = getattr(data, key)
            except AttributeError:
                if prefix:
                    msg = f"{type(self).__name__}() 'data' entry {prefix!r} has no attribute named {key}"
                else:
                    msg = f"{type(self).__name__}() 'data' has no attribute named {key}"
                raise AttributeError(msg)
        elif isinstance(data, (list, tuple)):
            try:
                index = int(key)
            except (TypeError, ValueError):
                if prefix:
                    msg = f"{type(self).__name__}() 'data' entry {prefix!r} is list or tuple, but key is no index"
                else:
                    msg = f"{type(self).__name__}() 'data' is list or tuple, but key is no index"
                raise AttributeError(msg)
            try:
                item = data[index]
            except IndexError:
                if prefix:
                    msg = f"{type(self).__name__}() 'data' entry {prefix!r} index {index} is out of bounds"
                else:
                    msg = f"{type(self).__name__}() 'data' index {index} is out of bounds"
                raise IndexError(msg)
        elif isinstance(data, Mapping):
            try:
                item = data[key]
            except KeyError:
                try:
                    index = int(key)
                    item = data[index]
                    key = index
                except (IndexError, KeyError, TypeError, ValueError):
                    if prefix:
                        msg = (
                            f"{type(self).__name__}() 'data' dict {prefix!r} must have key {key!r}"
                        )
                    else:
                        msg = f"{type(self).__name__}() 'data' dict must have key {key!r}"
                    raise KeyError(msg)
        else:
            if prefix:
                msg = f"{type(self).__name__}() 'data' entry {prefix!r} must be list, tuple, dict, dataclass, or namedtuple"
            else:
                msg = f"{type(self).__name__}() 'data' must be list, tuple, dict, dataclass, or namedtuple"
            raise TypeError(msg)
        # Transform item
        if subkey:
            item = self._apply_key(item, subkey, prefix=prefix + "." + key if prefix else key)
        else:
            item = self._apply_all(item)
        # Replace item, copy others if requested
        if is_dataclass(data):
            if self.copy:
                args = (
                    item if field.name == key else deepcopy(getattr(data, field.name))
                    for field in fields(data)
                )
                data = type(data)(*args)
            else:
                setattr(data, key, item)
        elif is_namedtuple(data):
            if self.copy:
                args = (item if k == key else deepcopy(v) for k, v in zip(data._fields, data))
                data = type(data)(*args)
            else:
                data = data._replace(**{key: item})
        elif isinstance(data, tuple):
            assert index is not None
            data = tuple(item if i == index else self._maybe_copy(v) for i, v in enumerate(data))
        elif isinstance(data, (list, tuple)):
            assert index is not None
            data = list(item if i == index else self._maybe_copy(v) for i, v in enumerate(data))
        else:
            assert isinstance(data, Mapping)
            data = {k: item if k == key else self._maybe_copy(v) for k, v in data.items()}
        return data

    def _maybe_copy(self, data: Any) -> Any:
        if self.copy:
            return deepcopy(data)
        return data

    def __repr__(self) -> str:
        return type(self).__name__ + f"({self.transform!r}, key={self.key!r}, copy={self.copy!r})"


class ItemwiseTransform:
    r"""Mix-in for data preprocessing and augmentation transforms."""

    @classmethod
    def item(
        cls,
        key: Union[int, str, KeysView, Iterable[Union[int, str]]],
        *args,
        ignore_meta: bool = True,
        ignore_missing: bool = False,
        **kwargs,
    ) -> ItemTransform:
        r"""Apply transform to specified item only."""
        return ItemTransform(
            cls(*args, **kwargs),
            key=key,
            ignore_meta=ignore_meta,
            ignore_missing=ignore_missing,
        )



r"""Metrics for evaluation of binary classification performance."""

from ignite.metrics.metrics_lambda import MetricsLambda
from ignite.metrics.confusion_matrix import ConfusionMatrix

import torch


EPSILON = 1e-15


def DiceCoefficient(cm: ConfusionMatrix) -> MetricsLambda:
    r"""Calculates Dice similarity coefficient from ``ignite.metrics.ConfusionMatrix``.

    Args:
        cm: Instance of confusion matrix metric, where
            ``cm[0, 0]=TN``, ``cm[0, 1]=FP``,
            ``cm[1, 0]=FN``, ``cm[1, 1]=TP``.

    Returns:
        MetricsLambda instance computing the Dice coefficient.

    """
    cm = cm.type(torch.DoubleTensor)  # Increase floating point precision
    return (2 * cm[1, 1] + EPSILON) / (2 * cm[1, 1] + cm[0, 1] + cm[1, 0] + EPSILON)


def FalseNegativeRate(cm: ConfusionMatrix) -> MetricsLambda:
    r"""Calculates miss rate from ``ignite.metrics.ConfusionMatrix``.

    Args:
        cm: Instance of confusion matrix metric, where
            ``cm[0, 0]=TN``, ``cm[0, 1]=FP``,
            ``cm[1, 0]=FN``, ``cm[1, 1]=TP``.

    Returns:
        MetricsLambda instance computing the true negative rate.

    """
    cm = cm.type(torch.DoubleTensor)  # Increase floating point precision
    return cm[1, 0] / (cm[1, 0] + cm[1, 1] + EPSILON)


def TrueNegativeRate(cm: ConfusionMatrix) -> MetricsLambda:
    r"""Calculates specificity from ``ignite.metrics.ConfusionMatrix``.

    Args:
        cm: Instance of confusion matrix metric, where
            ``cm[0, 0]=TN``, ``cm[0, 1]=FP``,
            ``cm[1, 0]=FN``, ``cm[1, 1]=TP``.

    Returns:
        MetricsLambda instance computing the true negative rate.

    """
    cm = cm.type(torch.DoubleTensor)  # Increase floating point precision
    return cm[0, 0] / (cm[0, 0] + cm[0, 1] + EPSILON)


def FalsePositiveRate(cm: ConfusionMatrix) -> MetricsLambda:
    r"""Calculates fall-out from ``ignite.metrics.ConfusionMatrix``.

    Args:
        cm: Instance of confusion matrix metric, where
            ``cm[0, 0]=TN``, ``cm[0, 1]=FP``,
            ``cm[1, 0]=FN``, ``cm[1, 1]=TP``.

    Returns:
        MetricsLambda instance computing the true positive rate.

    """
    cm = cm.type(torch.DoubleTensor)  # Increase floating point precision
    return cm[0, 1] / (cm[0, 1] + cm[0, 0] + EPSILON)


def TruePositiveRate(cm: ConfusionMatrix) -> MetricsLambda:
    r"""Calculates sensitivity from ``ignite.metrics.ConfusionMatrix``.

    Args:
        cm: Instance of confusion matrix metric, where
            ``cm[0, 0]=TN``, ``cm[0, 1]=FP``,
            ``cm[1, 0]=FN``, ``cm[1, 1]=TP``.

    Returns:
        MetricsLambda instance computing the true positive rate.

    """
    cm = cm.type(torch.DoubleTensor)  # Increase floating point precision
    return cm[1, 1] / (cm[1, 0] + cm[1, 1] + EPSILON)


def PositivePredictiveValue(cm: ConfusionMatrix) -> MetricsLambda:
    r"""Calculates precision from ``ignite.metrics.ConfusionMatrix``.

    Args:
        cm: Instance of confusion matrix metric, where
            ``cm[0, 0]=TN``, ``cm[0, 1]=FP``,
            ``cm[1, 0]=FN``, ``cm[1, 1]=TP``.

    Returns:
        MetricsLambda instance computing the positive predictive value.

    """
    cm = cm.type(torch.DoubleTensor)  # Increase floating point precision
    return cm[1, 1] / (cm[0, 1] + cm[1, 1] + EPSILON)


def NegativePredictiveValue(cm: ConfusionMatrix) -> MetricsLambda:
    r"""Calculates negative predictive value from ``ignite.metrics.ConfusionMatrix``.

    Args:
        cm: Instance of confusion matrix metric, where
            ``cm[0, 0]=TN``, ``cm[0, 1]=FP``,
            ``cm[1, 0]=FN``, ``cm[1, 1]=TP``.

    Returns:
        MetricsLambda instance computing the negative predictive value.

    """
    cm = cm.type(torch.DoubleTensor)  # Increase floating point precision
    return cm[0, 0] / (cm[0, 0] + cm[1, 0] + EPSILON)


Precision = PositivePredictiveValue
Recall = TruePositiveRate
Sensitivity = TruePositiveRate
Specificity = TrueNegativeRate
